/***************************************************************************/
/* The Minstrel's Song is (c) 1998-2005 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.                 */
/***************************************************************************/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "tms.h"

#ifdef CODE_SEGMENT
#pragma segment ACTION2_C
#endif

char *item_or_door_choice[2] =
    {
    "An item",
    "A door"
    };

char *poison_choice[5] =
    {
    "Weak poison",
    "Moderate poison",
    "Strong poison",
    "Very strong poison",
    "Extremely strong poison"
    };

long attempt_move(struct small_creature *c, long x, long y)
    {
    long x_target, y_target, i, j, best_distance, best_x, best_y,
      curr_distance;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    best_x = best_y = 0;
    x_target=y_target=0;
    if (x < c->x)
        x_target = -1;
    if (x > c->x)
        x_target = 1;
    if (y < c->y)
        y_target = -1;
    if (y > c->y)
        y_target = 1;
    best_distance = MAX_X + MAX_Y;
    if ((abs(c->x - x) > 1) || (abs(c->y - y) > 1))
        if ((flag_check(current_level->space[c->x + x_target][c->y + y_target]
          .flags, OCCUPIED)) || (!(flag_check(current_level->space[c->x
          + x_target][c->y + y_target].flags, CAN_OCCUPY))) ||
          ((flag_check(c->flags, WATER_ONLY)) &&
          (!(flag_check(current_level->space[c->x + x_target][c->y + y_target].
          flags, IS_WATER)))))
            {
            for(i = max(0, c->x + x_target - 1); i < min(MAX_X, c->x + x_target
              + 2); ++i)
                for(j = max(0, c->y + y_target - 1); j < min(MAX_Y, c->y +
                  y_target + 2); ++j)
                    {
                    curr_distance = abs(c->x - x) + abs(c->y - y);
                    if (!((flag_check(current_level->space[c->x + x_target][c->y
                      + y_target].flags, OCCUPIED)) || (!(flag_check(
                      current_level->space[c->x + x_target][c->y + y_target].
                      flags, CAN_OCCUPY))) || ((flag_check(c->flags,
                      WATER_ONLY)) && (!(flag_check(current_level->space[c->x
                      + x_target][c->y + y_target].flags, IS_WATER))))))
                        if (curr_distance < best_distance)
                            {
                            best_distance = curr_distance;
                            best_x = i;
                            best_y = j;
                            }
                    }
            x_target = best_x;
            y_target = best_y;
            }
    return move_creature(c, x_target, y_target);
    }

void dismount(struct small_creature *c)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (c != NULL)
        if (c->extended_creature != NULL)
            if (c->extended_creature->mount != NULL)
                {
                c->extended_creature->mount->x = c->x;
                c->extended_creature->mount->y = c->y;
                no_hostile_add_creature(current_level, c->extended_creature->mount,
                  c->x, c->y);
                c->extended_creature->mount=NULL;
                }
    }

void examine(struct small_creature *c)
    {
    long x, y;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (loc_xy(c, &x, &y))
        describe_space(c, x, y);
    }

void role_play_info(void)
    {
    window_clear();
    game_variables->fs_messages_printed = 0;
    fs_pmsg("The Minstrel's Song, as well as being a computer game, is a tabletop role");
    fs_pmsg("playing game developed by Jonathan Hayward.  Please select which document");
    fs_pmsg("you would like to see.  The cultures description, as well as telling about role");
    fs_pmsg("play, also gives background for the world in which this game.");
    fs_pmsg("");
    fs_pmsg("a: Espiriticthus: Cultures of a Fantasy World not Touched by Sin");
    fs_pmsg("b: The Minstrel's Song: Player's Introduction");
    fs_pmsg("c: The Minstrel's Song: Game Master's Introduction");
    fs_pmsg("d: The Minstrel's Song: Why Play?");
    fs_pmsg("e: The Minstrel's Song: Additional Notes");
    fs_pmsg("f: The Minstrel's Song: a Complex Mathematical Model (optional)");
    fs_pmsg("f: The Minstrel's Song: a Simplified Mathematical Model (optional)");
    fs_pmsg("");
    fs_pmsg("Your selection:");
    /* ## */
    switch(tgetc())
        {
        case 'a': read_file(FILE_CULTURES); break;
        case 'b': read_file(FILE_PLAYER_INTRO); break;
        case 'c': read_file(FILE_GM_INTRO); break;
        case 'd': read_file(FILE_WHYPLAY); break;
        case 'e': read_file(FILE_NOTES); break;
        case 'f': read_file(FILE_MODEL); break;
        case 'g': read_file(FILE_SIMPLE_MODEL); break;
        default: window_clear(); break;
        }
        game_variables->fs_messages_printed = 0;
    refresh_screen();
    }

/*
 *
 * Have a creature's mount act.  Now, this is limited to attacking any adjacent
 * creatures to which it is hostile, and having the creature delay if no
 * attack is made.
 *
 */

void mount_act(struct small_creature *c)
    {
    long i, j, my_continue;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    my_continue = TRUE;
    if (c == NULL || c->extended_creature == NULL ||
      c->extended_creature->mount == NULL)
        return;
    if (!(c->extended_creature == NULL))
        {
        if (is_hostile(c->extended_creature->mount, c))
            {
            reset_hostility(c->extended_creature->mount, c);
            }
        if (!is_following(c->extended_creature->mount, c))
            {
            make_follow(c->extended_creature->mount, c);
            }
        if (!(c->extended_creature->mount == NULL))
            {
            for(i=max(0, c->x - 1); i < min(MAX_X, c->x + 2); ++i)
                for(j = max(0, c->y - 1); j < min(MAX_Y, c->y + 2); ++j)
                    if (my_continue)
                        if (!(current_level->space[i][j].creatures == NULL))
                            if (is_hostile(c->extended_creature->mount,
                              current_level->space[i][j].creatures->current))
                                {
                                my_continue = FALSE;
                                attack_hth(c->extended_creature->mount,
                                  current_level->space[i][j].creatures->current
                                  );
                                }
            if (my_continue) /* Creature hasn't attacked anything */
            c->extended_creature->mount->nextaction += 10 * current_level->timescale;
            }
        }
    }

void give_name(struct small_creature *namer, struct small_creature *ctarget)
    {
    char message[256], this_name[256];
    struct small_creature *dummy;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) namer);
    check_pointer((struct pointer_struct *) ctarget);
#endif
    if (!(ctarget == NULL))
        {
        find_title(ctarget, this_name);
#ifdef HAS_SNPRINTF
        snprintf(message, 255, "What do you want to name %s?", this_name);
#else
        sprintf(message, "What do you want to name %s?", this_name);
#endif
        message[255] = '\0';
#ifdef GTK
        if (new_tgets(message, this_name, NULL))
#else
        pmsg(message);
        if (tgets(this_name))
#endif
            {
            if (flag_check(ctarget->flags, HUMAN))
#ifdef HAS_SNPRINTF
                snprintf(message, 255,
                  "And do you want me to still tell you %s's race?", this_name);
#else
                sprintf(message, 
                  "And do you want me to still tell you %s's race?", this_name);
#endif
            else
#ifdef HAS_SNPRINTF
                snprintf(message, 255,
                  "And do you want me to still tell you what kind of animal %s is?", this_name);
#else
                sprintf(message,
                  "And do you want me to still tell you what kind of animal %s is?", this_name);
#endif
            message[255] = '\0';
#ifdef GTK
            if (tms_confirm(message))
#else
            pmsg(message);
            if (confirm_silent())
#endif
                {
                strncat(this_name, " the ", max(0, 255 - strlen(this_name)));
                this_name[255] = '\0';
                strncat(this_name, master_creatures[ctarget->identity]->species, max(0, 255 - strlen(this_name)));
                this_name[255] = '\0';
                }
            ctarget->name = (struct text_struct *) calloc(1, sizeof(struct
              text_struct));
            ctarget->name->me = ctarget->name;
            strncpy(ctarget->name->text, this_name, 117);
            ctarget->name->text[117] = '\0';
            ctarget->name->next = NULL;
            }
        else
            pmsg("Aborted.");
        }
    dummy = namer;
    }

void auto_move(struct small_creature *c, long x_direction, long y_direction,
  long flags)
    {
    long i, j;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if ((c->x + x_direction < 0) || (c->x + y_direction >= MAX_X) ||
      (c->y + y_direction < 0) || (c->y + y_direction >= MAX_Y) ||
      ((x_direction == 0) && (y_direction == 0)))
        flag_reset(&c->flags, flags);
    else if ((flag_check(current_level->space[c->x + x_direction]
      [c->y + y_direction].flags, OCCUPIED)) || (!(flag_check(current_level->
      space[c->x + x_direction][c->y + y_direction].flags, CAN_OCCUPY))))
        flag_reset(&c->flags, flags);
    else if (flag_check(flags, CONTROL_MOVE))
        for(i = max(0, c->x - 1); i < min(c->x + 2, MAX_X); ++i)
            for(j = max(0, c->y - 1); j < min(c->y + 2, MAX_Y); ++j)
                if (((current_level->space[i][j].property != NONE) &&
                  (current_level->space[i][j].property != P_SECRET)) ||
                  (current_level->space[i][j].items != NULL) ||
                  (current_level->space[i][j].type == L_DOOR) ||
                  ((current_level->space[i][j].creatures != NULL) &&
                  ((i != c->x) || (j != c->y))) ||
                  (current_level->space[i][j].au > 0) ||
                  (current_level->space[i][j].el > 0) ||
                  (current_level->space[i][j].si > 0) ||
                  (current_level->space[i][j].cu > 0))
                    flag_reset(&c->flags, flags);
    if (flag_check(c->flags, flags))
        {
        move_creature(c, x_direction, y_direction);
        if ((c == player) && (flag_check(game_variables->flags, MAX_REFRESH)))
            update_window();
        }
    }

void rest(struct small_creature *c)
    {
    long amount, safe, success;
    struct clist *traverse;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    safe = TRUE;
    for(traverse = current_level->creatures; traverse != NULL; traverse =
      traverse->next)
        if (traverse->current != NULL)
        if (is_hostile(traverse->current, c))
            safe = FALSE;
    if (!(safe))
        {
#ifdef GTK
        success = tms_confirm_negative("It is not safe to rest here. Rest anyway?");
#else
        pmsg("It is not safe to rest here.  Rest anyway?");
        success = confirm_silent();
#endif
        }
    else
        success = TRUE;
    if (success)
        {
        amount = time_parse();
        pass_time(amount);
        }
    }

void do_action(struct small_creature *c)
    {
    char player_input[256];
    long which_action, value;
    struct small_item *i1, *i2;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    which_action = action_parse();
    switch(which_action)
        {
        case ACTION_BECOME_BELLICOSE:
            flag_set(&c->flags, BELLICOSE);
        break;
        case ACTION_CHARITY_CONTACTS:
            read_file(FILE_CHARITY_CONTACTS);
        break;
        case ACTION_CLEAN:
            if (c->extended_creature != NULL)
                append_ilist(&c->extended_creature->pack,
                  clean(get_item(c, EXTRACT_REMOVE_ITEM|EXTRACT_RECURSIVE)));
        break;
        case ACTION_COVER_WITH_POISON:
            if (c->extended_creature != NULL)
                cover_with_poison(c, get_item(c, EXTRACT_RECURSIVE|
                  EXTRACT_PROMPT_BOTH_PACK_AND_EQUIPMENT));
        break;
        case ACTION_DEBUG:
            debug();
        break;
        case ACTION_DISMOUNT:
            pmsg("You dismount...");
            dismount(c);
        break;
        case ACTION_DROP_MONEY:
            drop_money(c);
        break;
        case ACTION_EXTENDED_HELP:
            read_file(FILE_EXTENDED_HELP);
        break;
        case ACTION_FILL_WITH_WATER:
            fill_with_water(get_item(c, EXTRACT_RECURSIVE));
            calc_ready(c);
        break;
        case ACTION_FILL_WITH_OIL:
            fill_with_oil(c);
        break;
        case ACTION_GATHER_FROM_TREE:
            gather_from_tree(c);
        break;
        case ACTION_GIVE_NAME:
            give_name(c, loc_creature(c));
        break;
        case ACTION_HASTE:
#ifdef GTK
            if (new_tgets("What haste value do you want (-10 to 50)?",
              player_input, NULL))
#else
            pmsg("What haste value do you want (-10 to 50)?");
            if (tgets(player_input))
#endif
                {
#ifdef MAC_68k
                value = strtol(player_input, NULL, 10);
#else
                sscanf(player_input, "%ld", &value);
#endif
                value = min(value, 50);
                value = max(value, -10);
                c->extended_creature->haste = value;
                }
        break;
        case ACTION_INSERT:
            if (c->extended_creature != NULL)
                {
                pmsg("Item to be inserted:");
                i1 = get_item(c, EXTRACT_RECURSIVE|EXTRACT_REMOVE_ITEM);
                pmsg("Item to insert it into:");
                i2 = get_item(c, EXTRACT_RECURSIVE);
                if ((i2 != NULL) && (i1 != NULL))
                    {
                    if (!(my_insert(current_level, i1, i2)))
                        {
                        pmsg("That won't go in.");
                        add_item(&c->extended_creature->pack, i1);
                        }
                    }
                }
        break;
        case ACTION_LIST_CHANGES:
            read_file(FILE_CHANGES);
        break;
        case ACTION_MOUNT:
            attempt_mount(c, loc_creature(c));
        break;
        case ACTION_NAP:
            if (c != NULL)
                {
                go_to_sleep(c);
                if (c->extended_creature != NULL)
                    {
                    c->extended_creature->stati->time[STATUS_SLEEPING] =
                      c->extended_creature->stati->time[STATUS_TIRED] / 3;
                    }
                }
        break;
        case ACTION_PET:
            pet(c, loc_creature(c));
        break;
        case ACTION_PRINT_SKILL_VALUES:
            print_attributes(c);
        break;
        case ACTION_PUSH_CREATURE:
            push_creature(c);
        break;
        case ACTION_READ_FAQ_LIST:
            read_file(FILE_FAQ_LIST);
        break;
        case ACTION_REMOVE:
            pmsg("Remove which item from container?");
            i1 = get_item(c, EXTRACT_RECURSIVE|EXTRACT_REMOVE_ITEM);
            if (i1 != NULL)
                add_item(&c->extended_creature->pack, i1);
        break;
        case ACTION_RUN:
            flag_set(&c->flags, RUNNING);
            pmsg("You are now running.");
        break;
        case ACTION_SET_OPTIONS:
            set_options();
        break;
        case ACTION_SLEEP_UNTIL_SUNRISE:
            pmsg("You lie down, to sleep until the sun rises.");
            go_to_sleep(player);
            player->extended_creature->stati->time[STATUS_SLEEPING] = DAY / 2;
            flag_set(&game_variables->flags, PLAYER_SLEEPING_UNTIL_SUNRISE);
            while ((!(flag_check(game_variables->flags, DAYLIGHT))) &&
              (status_check(player, STATUS_SLEEPING) > 0))
                {
                pass_time(MINUTE);
                update_light_sources();
                }
        break;
        case ACTION_SWAP_PLACES:
            swap_places(c, one_move_loc_creature(c));
        break;
        case ACTION_WALK:
            flag_reset(&c->flags, RUNNING);
            pmsg("You are now walking.");
        break;
        }
    }

void pet(struct small_creature *petter, struct small_creature *pettee)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) petter);
    check_pointer((struct pointer_struct *) pettee);
#endif
    if ((petter != NULL) && (pettee != NULL))
        {
        if (!(flag_check(pettee->flags, HUMAN)))
            {
            switch(trandom() % 4)
                {
                case 0: case 1: case 2:
                    pmsg("Pat, pat, pat.");
                break;
                case 3:
                    pmsg("You give a nice scratch behind the ears.");
                break;
                }
            if (pettee->identity == DOLPHIN)
                {
                flag_set(&game_variables->quests_completed, PETTED_DOLPHIN);
                pmsg("The dolphin nuzzles against you.");
                }
            }
        else
            pmsg("Hugs, kisses, etc. are given as the greeting within the talk command.");
        }
    }

void close_door(struct small_creature *c)
    {
    long x, y;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    pmsg("Close a door:");
    if (loc_xy(c, &x, &y))
        {
        if ((abs(c->x - x) <= 1) && (abs(c->y - y) <= 1))
            {
            if ((current_level->space[x][y].type == L_DOOR) ||
              (current_level->space[x][y].type == L_CLOCKWORK_DOOR))
                {
                if (!(flag_check(current_level->space[x][y].flags,
                  DOOR_CLOSED)))
                    {
                    flag_set(&current_level->space[x][y].flags, DOOR_CLOSED);
                    flag_reset(&current_level->space[x][y].flags, CAN_OCCUPY);
                    current_level->space[x][y].icon = ICON_CLOSED_DOOR;
                    }
                else
                    pmsg("The door is already closed.");
                }
            else
                pmsg("That is not a door.");
            }
        else
            pmsg("You need to be next to a door to close it.");
        }
    }

void open_door(struct small_creature *c)
    {
    long x, y;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    pmsg("Open a door:");
    if (loc_xy(c, &x, &y))
        {
        if ((abs(c->x - x) <= 1) && (abs(c->y - y) <= 1))
            {
            if ((current_level->space[x][y].type == L_DOOR) ||
              (current_level->space[x][y].type == L_CLOCKWORK_DOOR))
                {
                if (flag_check(current_level->space[x][y].flags,
                  DOOR_CLOSED))
                    {
                    if (!(flag_check(current_level->space[x][y].flags,
                    DOOR_LOCKED)))
                        {
                        flag_reset(&current_level->space[x][y].flags,
                          DOOR_CLOSED);
                        flag_set(&current_level->space[x][y].flags,
                          CAN_OCCUPY);
                        current_level->space[x][y].icon = ICON_OPEN_DOOR;
                        }
                    else
                        {
                        pmsg("The door appears to be locked.");
                        if (current_level->space[x][y].type ==
                          L_CLOCKWORK_DOOR)
                            {
#ifdef GTK
                            if (tms_confirm("It is a clockwork door; do you wish to attempt to figure it out?"))
#else
                            pmsg("It is a clockwork door; do you wish to attempt to figure it out?");
                            if (confirm_silent())
#endif
                                {
                                if (p_puzzle_door(c, x, y))
                                    {
                                    flag_reset(&current_level->space[x][y].flags, DOOR_LOCKED);
                                    }
                                }
                            else
                                pmsg("You back away...");
                            }
                        }
                    }
                else
                    pmsg("The door is already open.");
                }
            else
                pmsg("That is not a door.");
            }
        else
            pmsg("You need to be next to a door to open it.");
        }
    }

void city_move(struct small_creature *c)
    {
    long i, which;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (flag_check(current_level->flags, IS_CITY))
        {
        which = location_parse();
        if (which != NONE)
            {
            if (which == EXIT_SETTLEMENT)
                move_creature(c, 0 - c->x, 0 - c->y);
            else
                for(i = 0; i < NUM_LOCATION_TYPES; ++i)
                    if (current_level->location[i].identity == which)
                        {
                        if (flag_check(current_level->location[i].flags,
                          LOCATION_KNOWN))
                            {
                            if (current_level->location[i].x == c->x &&
                              current_level->location[i].y == c->y)
                                pmsg("You're already there!");
                            else
                                move_creature(c, current_level->location[i].x -
                                  c->x, current_level->location[i].y - c->y);
                            }
                        else
                            pmsg("You don't know where that is.");
                        }
            }
        }
    else
        pmsg("You need to be in a city to do this command.");
    }

void e_horn(struct small_creature *c)
    {
    char creature_title[256], message[256];
    struct clist *traverse;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (c == player)
        pmsg("You blow it...");
    else
        if (can_see(player, c->x, c->y))
            {
            find_title(c, creature_title);
#ifdef HAS_SNPRINTF
            snprintf(message, 255, "%s blows a horn!", creature_title);
#else
            sprintf(message, "%s blows a horn!", creature_title);
#endif
            message[255] = '\0';
            pmsg(message);
            }
    if (current_level->creatures != NULL)
        for(traverse = current_level->creatures; traverse != NULL; traverse =
          traverse->next)
        if (traverse->current != NULL)
            flag_reset(&traverse->current->flags, INACTIVE);
    }

void aid_open(struct small_creature *c)
    {
    char response;
    long x, y;
    struct small_item *this_item;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
#ifdef GTK
    response = menu_choice("Attempt to open:", item_or_door_choice, 2);
    if (response == 0)
        response = 'i';
    else if (response == 1)
        response = 'd';
    switch(response)
#else
    pmsg("Attempt to open (i)tem, or (d)oor?");
    switch(tgetc())
#endif
        {
        case 'i':
            this_item = get_item(c, EXTRACT_RECURSIVE);
            if (this_item != NULL)
                {
                c->nextaction += MINUTE;
                if (c->extended_creature != NULL)
                    {
                    if (skill_check(c, c->extended_creature->attributes[ST], DIFFICULT,
                      SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE))
                        {
                        pmsg("You open it.");
                        flag_set(&this_item->flags, IS_OPEN);
                        }
                    else
                        pmsg("That didn't work.");
                    }
                }
        break;
        case 'd':
            pmsg("Which door?");
            if (loc_xy(c, &x, &y))
                {
                if ((current_level->space[x][y].type == L_DOOR) ||
                  (current_level->space[x][y].type == L_CLOCKWORK_DOOR))
                    {
                    c->nextaction += MINUTE;
                    if (flag_check(current_level->space[x][y].flags,
                      DOOR_CLOSED))
                        {
                        if (c->extended_creature != NULL)
                            {
                            if (skill_check(c, c->extended_creature->attributes[ST],
                              DIFFICULT,
                              SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE))
                                {
                                pmsg("You force the door open!");
                                flag_reset(&current_level->space[x][y].flags,
                                  DOOR_CLOSED);
                                flag_set(&current_level->space[x][y].flags,
                                  CAN_OCCUPY);
                                current_level->space[x][y].icon =
                                  ICON_OPEN_DOOR;
                                }
                            else
                                pmsg("You can't seem to open it.");
                            }
                        }
                    else
                        pmsg("That door is already open!");
                    }
                else
                    pmsg("That is not a door.");
                }
        break;
        }
    }

void attempt_mount(struct small_creature *c1, struct small_creature *c2)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c1);
    check_pointer((struct pointer_struct *) c2);
#endif
    if ((c1->extended_creature != NULL) && (c2 != NULL))
        {
        if (is_following(c2, c1))
            {
            if (c1->extended_creature->natural_weight <
            master_creatures[c2->identity]->
              natural_weight && master_creatures[c2->identity]->natural_weight
              / 3)
                {
                c1->extended_creature->mount = c2;
                remove_creature(current_level, c2);
                pmsg("You are now mounted!");
                /*
                if (c2->identity == DRAGON)
                    flag_set(&game_variables->quests_completed, RODE_DRAGON);
                */
                }
            else
                pmsg("You are too heavy for such a creature as mount.");
            }
        else
            pmsg("That creature is not friendly to you.");
        }
    }

void ignite(struct small_creature *c, struct small_item *i)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
    check_pointer((struct pointer_struct *) i);
#endif
    if (i != NULL)
        {
        if (flag_check(i->flags, FLAMMABLE))
            {
            pmsg("You light it...");
            switch(orig_items[i->identity]->effect)
                {
                case GIVES_LIGHT:
                    pmsg("It gives light around you.");
                    flag_set(&i->flags, ACTIVATED);
                    add_new_light_source_to_creature(c,
                      orig_items[i->identity]->range);
                    flag_set(&c->flags, HAS_LIGHT_SOURCE);
                    update_light_sources();
                    morewait();
                break;
                default:
                    pmsg("It burns and gives you warmth.");
                    status_add(c, STATUS_COLD, 0 - 1 * (1 + (i->weight / LB)));
                    if (c->extended_creature != NULL)
                        {
                        find_and_remove_item(i, &c->extended_creature->pack);
                        find_and_remove_item(i, &c->extended_creature->wear_wield);
                        }
                break;
                }
            }
        }
    }

void fill_with_oil(struct small_creature *c)
    {
    struct small_item *this_item;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    pmsg("Select a light source to fill with oil:");
    this_item = get_item(c, EXTRACT_RECURSIVE);
    if ((this_item != NULL) && (c->extended_creature != NULL))
        {
        if ((this_item->identity == LANTERN) || (this_item->identity ==
          STAFF_LANTERN) || (this_item->identity == FIRESTAR))
            {
            if (find_and_remove_id(&c->extended_creature->pack, OIL))
                {
                pmsg("You fill it with oil.");
                this_item->uses = orig_items[this_item->identity]->maxuses;
                if (this_item->identity == FIRESTAR)
                    this_item->uses = orig_items[this_item->identity]->maxuses;
                }
            else
                pmsg("You don't have any oil!");
            }
        else
            pmsg("I don't see the point of filling that with oil.");
        }
    }

void cover_with_poison(struct small_creature *c, struct small_item *i)
    {
    char player_response;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
    check_pointer((struct pointer_struct *) i);
#endif
    if ((c->extended_creature != NULL) && (i != NULL))
        {
        if (flag_check(i->flags2, IS_WEAPON))
            {
            if (!(flag_check(i->flags2, I_MISSLE_WEAPON)))
                {
                if (!(flag_check(i->flags2, I_BLUNT_DAMAGE)))
                    {
                    if (i->uses == 0)
                        {
#ifdef GTK
                        player_response = menu_choice("Apply which strength of poison to weapon: ", poison_choice, 5);
                        if (player_response == 0)
                            player_response = 'w';
                        else if (player_response == 1)
                            player_response = 'm';
                        else if (player_response == 2)
                            player_response = 's';
                        else if (player_response == 3)
                            player_response = 'v';
                        else if (player_response == 4)
                            player_response = 'e';
                        switch(player_response)
#else
                        pmsg("Apply poison to weapon: (w)eak, (m)oderate, (s)trong, (v)ery strong, or");
                        pmsg("(e)xtremely strong poison?");
                        switch(tgetc())
#endif
                            {
                            case 'w':
                                if (find_and_remove_id(&c->extended_creature->pack,
                                  POISON_WEAK))
                                    {
                                    i->other = 10;
                                    ++i->uses;
                                    pmsg("Your weapon is covered with poison.");
                                    }
                                else
                                    pmsg("You don't have any poison of that strength.");
                            break;
                            case 'm':
                                if (find_and_remove_id(&c->extended_creature->pack,
                                  POISON_MODERATE))
                                    {
                                    i->other = 20;
                                    ++i->uses;
                                    pmsg("Your weapon is covered with poison.");
                                    }
                                else
                                    pmsg("You don't have any poison of that strength.");
                            break;
                            case 's':
                                if (find_and_remove_id(&c->extended_creature->pack,
                                  POISON_STRONG))
                                    {
                                    i->other = 30;
                                    ++i->uses;
                                    pmsg("Your weapon is covered with poison.");
                                    }
                                else
                                    pmsg("You don't have any poison of that strength.");
                            break;
                            case 'v':
                                if (find_and_remove_id(&c->extended_creature->pack,
                                  POISON_VERY_STRONG))
                                    {
                                    i->other = 40;
                                    ++i->uses;
                                    pmsg("Your weapon is covered with poison.");
                                    }
                                else
                                    pmsg("You don't have any poison of that strength.");
                            break;
                            case 'e':
                                if (find_and_remove_id(&c->extended_creature->pack,
                                  POISON_EXTREMELY_STRONG))
                                    {
                                    i->other = 50;
                                    ++i->uses;
                                    pmsg("Your weapon is covered with poison.");
                                    }
                                else
                                    pmsg("You don't have any poison of that strength.");
                            break;
                            default:
                                pmsg("Aborted.");
                            break;
                            }
                        }
                    else
                        pmsg("That weapon is already poisoned.");
                    }
                else
                    pmsg("Why would you want to put poison on a blunt weapon?!?");
                }
            else
                pmsg("Only hand to hand weapons may be covered with poison.");
            }
        else
            pmsg("You may only cover weapons with poison.");
        }
    }

void janra_ball_internal(void)
    {
    window_clear();
    read_random_excerpt(FILE_JANRA_BALL);
    pass_time(MINUTE);
    fs_anykey();
    while (trandom() % 7 != 0)
        {
        window_clear();
        read_random_excerpt(FILE_JANRA_BALL);
        pass_time(MINUTE);
        fs_anykey();
        }
    if (player->identity != JANRA)
        pmsg("You have a splitting headache.");
    }

void e_janra_ball(struct small_creature *c)
    {
    struct small_creature *other_player;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (c->identity == JANRA)
        {
        pmsg("Play Janra ball with whom?");
        other_player = loc_creature(c);
        if (other_player != NULL)
            {
            if (other_player != c)
                {
                if (other_player->identity == JANRA)
                    janra_ball_internal();
                else
                    {
                    if (other_player->extended_creature != NULL)
                        switch(other_player->extended_creature->gender)
                            {
                            case MALE:
                                pmsg("He is not a Janra!");
                            break;
                            case FEMALE:
                                pmsg("She is not a Janra!");
                            break;
                            default:
                                pmsg("That is not a Janra!");
                            break;
                            }
                    else
                        pmsg("That is not a Janra!");
                    }
                }
            else
                pmsg("I meant play Janra ball with whom else?");
            }
        else
            pmsg("Aborted.");
        }
    else
        {
        pmsg("You have a splitting headache.");
        take_time(c, MINUTE);
        }
    }

void attempt_fly(struct small_creature *c)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (c != NULL)
        {
        if (!(flag_check(c->flags, ABOVE_GROUND)))
            {
            if (flag_check(current_level->flags, OUTDOORS))
                {
                if (can_fly(c))
                    {
                    flag_set(&c->flags, ABOVE_GROUND);
                    if (c == player)
                        pmsg("You are flying!");
                    }
                else
                    if (c == player)
                        pmsg("You can't fly.");
                }
            else
                if (c == player)
                    pmsg("You must be outdoors to fly!");
            }
        else
            if (c == player)
                pmsg("You are already flying!");
        }
    }

void land_from_flying(struct small_creature *c)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (c != NULL)
        {
        if (flag_check(c->flags, ABOVE_GROUND))
            {
            if (c == player)
                pmsg("You land on the ground.");
            }
        flag_reset(&c->flags, ABOVE_GROUND);
        }
    }

void action_fly(struct small_creature *c)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (c != NULL)
        {
        if (!(flag_check(c->flags, ABOVE_GROUND)))
            attempt_fly(c);
        else
            land_from_flying(c);
        }
    }

void gather_from_tree(struct small_creature *c)
    {
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (current_level->space[c->x][c->y].type == L_TREE)
        space_effect(c, c->x, c->y);
    else
        if (c == player)
            pmsg("You are not under a tree.");
    }

void write_on_roster(void)
    {
    struct text_struct *master, *traverse;
    char player_name[256], real_filename[256];
    FILE *hero_roster;
    master = NULL;
    append_text_to_text_struct(&master, "");
    append_text_to_text_struct(&master, "--");
    append_text_to_text_struct(&master, "");
    strncpy(real_filename, LIB_DIR, 255);
    real_filename[255] = '\0';
    strncat(real_filename, FILE_BOOK_ROSTER_OF_HEROES, max(0, 255 -
      strlen(LIB_DIR)));
    real_filename[255] = '\0';
#ifdef GTK
    new_tgets("What is your name?", player_name, NULL);
#else
    pmsg("What is your name?");
    tgets(player_name);
#endif
    append_text_to_text_struct(&master, player_name);
    append_text_to_text_struct(&master, "");
    write_on_item(&master);

/*
 * 
 * We don't worry about file locking because this will be an extremely rare
 * event, and almost instantaneous (the time for player writing having been
 * taken in write_on_item).
 *
 */

    hero_roster = fopen(real_filename, "a");
    if (hero_roster)
        {
        for(traverse = master; traverse != NULL; traverse = traverse->next)
            fprintf(hero_roster, "%s\n", traverse->text);
        fclose(hero_roster);
        }
    else
        {
        pmsg("I'm sorry; I wasn't able to open the file.  I'll read the text appended, so you");
        nocap_pmsg("can copy it yourself.");
        morewait();
        read_text_struct(master);
        morewait();
        morewait();
        }
    free_text_struct(master);
    }

void push_creature(struct small_creature *c)
    {
    long x, y;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    if (one_move(c, &x, &y))
        {
        if ((c->x + 2 * x >= 0) && (MAX_X > c->x + 2 * x) && (c->y + 2 * y >=
          0) && (MAX_Y > c->y + 2 * y))
            {
            if (current_level->space[c->x + x][c->y + y].creatures != NULL)
                {
                move_creature(current_level->space[c->x + x][c->y +
                  y].creatures->current, x, y);
                move_creature(c, x, y);
                }
            else
                pmsg("There isn't any creature there!");
            }
        else
            pmsg("You can't push a creature off the level!");
        }
    }

void drop_money(struct small_creature *c)
    {
    char choice, input[256];
    long amount, *coin1, *coin2;
#ifdef DEBUG
    check_pointer((struct pointer_struct *) c);
#endif
    coin1 = coin2 = NULL;
    if (c != NULL)
        if (c->extended_creature != NULL)
            {
#ifdef GTK
            choice = 'a';
            while ((choice != 'g') && (choice != 'e') && (choice != 's') &&
              (choice != 'c') && (choice != ESCAPE))
                choice = new_tgetc("Drop (g)old, (e)lectrum, (s)ilver, or (c)opper?");
#else
            pmsg("Drop (g)old, (e)lectrum, (s)ilver, or (c)opper?");
            choice = 'a';
            while ((choice != 'g') && (choice != 'e') && (choice != 's') &&
              (choice != 'c') && (choice != ESCAPE))
                choice = tgetc();
#endif
            if (choice != ESCAPE)
                {
#ifdef GTK
                new_tgets("How much?", input, NULL);
#else
                pmsg("How much?");
                tgets(input);
#endif
#ifdef MAC_68k
                amount = strtol(input, NULL, 10);
#else
                sscanf(input, "%ld", &amount);
#endif
                switch(choice)
                    {
                    case 'g':
                        coin1 = &c->extended_creature->au;
                        coin2 = &current_level->space[c->x][c->y].au;
                    break;
                    case 'e':
                        coin1 = &c->extended_creature->el;
                        coin2 = &current_level->space[c->x][c->y].el;
                    break;
                    case 's':
                        coin1 = &c->extended_creature->si;
                        coin2 = &current_level->space[c->x][c->y].si;
                    break;
                    case 'c':
                        coin1 = &c->extended_creature->cu;
                        coin2 = &current_level->space[c->x][c->y].cu;
                    break;
                    }
                if (amount <= 0)
                    pmsg("Aborted.");
                else if (*coin1 == 0)
                    pmsg("You don't have any of that kind of money.");
                else
                    {
                    if (amount > *coin1)
                        pmsg("You don't have that much; dropping what you have...");
                    amount = min(amount, *coin1);
                    *coin1 = max(0, *coin1 - amount);
                    *coin2 += amount;
                    }
                }
            else
                pmsg("Aborted.");
            }
    calc_ready(c);
    }

void write_on_log(void)
    {
    struct text_struct *master, *traverse;
    char real_filename[256];
    FILE *hero_roster;
    strncpy(real_filename, LIB_DIR, 255);
    real_filename[255] = '\0';
    strncat(real_filename, FILE_BOOK_THE_WRITINGS_OF_THE_ADVENTURERS, max(0,
      255 - strlen(LIB_DIR)));
    real_filename[255] = '\0';
    master = NULL;
    append_text_to_text_struct(&master, "");
    append_text_to_text_struct(&master, "--");
    append_text_to_text_struct(&master, "");
    write_on_item(&master);

/*
 * 
 * We don't worry about file locking because this will be an extremely rare
 * event, and almost instantaneous (the time for player writing having been
 * taken in write_on_item).
 *
 */

    hero_roster = fopen(real_filename, "a");
    if (hero_roster)
        {
        for(traverse = master; traverse != NULL; traverse = traverse->next)
            fprintf(hero_roster, "%s\n", traverse->text);
        fclose(hero_roster);
        }
    else
        {
        pmsg("I'm sorry; I wasn't able to open the file.  I'll read the text appended, so you");
        nocap_pmsg("can copy it yourself.");
        morewait();
        read_text_struct(master);
        morewait();
        morewait();
        }
    free_text_struct(master);
    }
