/***************************************************************************/
/* The Minstrel's Song is (c) 1998-2005 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.                 */
/***************************************************************************/
#include "tms.h"

#ifdef CODE_SEGMENT
#pragma segment COLLECTION_C
#endif

/* Garbage collection routines.  */

/* Free a list of items, but not the items themselves.  */

void free_ilist(struct ilist *il)
    {
    if (il == NULL || il->me != il)
        return;
    if (!(il->next == NULL))
        free_ilist(il->next);
    il->me = NULL;
    free(il);
    }

/* Free a list of creatures, but not the creatures themselves.  */

void free_clist(struct clist *cl)
    {
    if (cl == NULL || cl->me != cl)
        return;
    if (cl != NULL)
        {
        if (!(cl->next == NULL))
            free_clist(cl->next);
        cl->me = NULL;
        free(cl);
        }
    }

/* Free a list of items, including the items themselves.  */

void free_items(struct ilist *il)
    {
    if (il == NULL || il->me != il)
        return;
    if (!(il->next == NULL))
        free_items(il->next);
    free_item(il->current);
    il->me = NULL;
    free(il);
    }

/* Free a list of creatures, including the creatures themselves.  */

void free_creatures(struct clist *cl)
    {
    if (cl == NULL || cl->me != cl)
        return;
    if (!(cl->next == NULL))
        free_creatures(cl->next);
    free_creature(cl->current);
    cl->me = NULL;
    free(cl);
    }

/* Free a list of light sources, but not their holders.  */

void free_light_source(struct light_source_struct *l)
    {
    if (l == NULL || l->me != l)
        return;
    if (!(l->next == NULL))
        free_light_source(l->next);
    l->me = NULL;
    free(l);
    }

/* Free an item and its contents.  */

void free_item(struct small_item *i)
    {
    struct ilist *traverse, *previous, *curr_ilist;
    if (i == NULL || i->me != i)
        return;
    if (!(i->contents == NULL))
        free_items(i->contents);
    if (!(master_ilist == NULL))
        {
        if (master_ilist->current == i)
            {
            curr_ilist = master_ilist;
            master_ilist = master_ilist->next;
            curr_ilist->me = NULL;
            free(curr_ilist);
            }
        else
            for(traverse = previous = master_ilist; (traverse!=NULL);
              traverse = traverse->next)
        if (traverse->current != NULL)
                {
                if (traverse->current == i)
                    {
                    previous->next = traverse->next;
                    traverse->me = NULL;
                    free(traverse);
                    traverse = previous;
                    }
                previous = traverse;
                }
        }
    /* Commented out because this was triggering crashing. */
    /*
    if (i->creatures != NULL)
        {
        free_creatures(i->creatures);
        i->creatures = NULL;
        }
    */
    i->me = NULL;
    free(i);
    }

/* Free a creature, its possessions, and its mount, but not any followers,
hostile creatures, or leaders.  */

void free_creature(struct small_creature *c)
    {
    struct clist *traverse, *previous, *curr_clist;
    if (c == NULL || c->me != c)
        return;
    if (!(c->follow == NULL))
        free_clist(c->follow);
    if (!(c->hostile == NULL))
        free_clist(c->hostile);
    if (!(c->followed_by == NULL))
        free_clist(c->followed_by);
    if (!(c->extended_creature == NULL))
        free_master_creature(c->extended_creature);
    if (!(master_clist == NULL))
        {
        if (master_clist->current == c)
            {
            curr_clist = master_clist;
            master_clist = master_clist->next;
            curr_clist->me = NULL;
            free(curr_clist);
            }
        else
            {
            for (previous = traverse = master_clist; (!(traverse == NULL));
              traverse = traverse->next)
        if (traverse->current != NULL)
        {
                if (traverse->current == c)
                    {
                    previous->next = traverse->next;
                    traverse->me = NULL;
                    free(traverse);
                    traverse = previous;
                    }
                else
                    previous = traverse;
        }
            }
        }
    c->me = NULL;
    free(c);
    }

void free_faith_struct(struct faith_struct *f)
    {
    if (f == NULL || f->me != f)
        return;
    if (f->breath_prayer != NULL)
        free_text_struct(f->breath_prayer);
    f->me = NULL;
    free(f);
    }

void free_levels(struct llist *levels)
    {
    if (levels == NULL || levels->me != levels)
        return;
    if (!(levels->next == NULL))
        free_levels(levels->next);
    free_level(levels->current);
    levels->me = NULL;
    free(levels);
    }

void free_master_creature(struct master_creature *c)
    {
    if (!(c->mount == NULL))
        free_creature(c->mount);
    if (!(c->wear_wield == NULL))
        free_items(c->wear_wield);
    if (!(c->pack == NULL))
        free_items(c->pack);
    if (!(c->stati == NULL))
        free(c->stati);
    if (!(c->skills == NULL))
        free(c->skills);
    if (c->faith != NULL)
        free_faith_struct(c->faith);
    free(c);
    }

void free_shop(struct shop_struct *s)
    {
    if (s == NULL || s->me != s)
        return;
    if (!(s->items==NULL))
        free_items(s->items);
    if (!(s->creatures==NULL))
        free_creatures(s->creatures);
    s->me = NULL;
    free(s);
    }

/* We do not free the level associated with a special.  */

void free_special(struct special_struct *s)
    {
    if (s == NULL || s->me != s)
        return;
    if (!(s->shop==NULL))
        free_shop(s->shop);
    s->me = NULL;
    free(s);
    }

void free_text_struct(struct text_struct *t)
    {
    if (t == NULL || t->me != t)
        return;
    if (!(t->next == NULL))
        free_text_struct(t->next);
    t->me = NULL;
    free(t);
    }

void free_level(struct level *l)
    {
    long i, j;
    struct llist *this_llist, *traverse, *previous;
    if (l == NULL || l->me != l)
        return;
    if (!(l->light_sources==NULL))
        free_light_source(l->light_sources);
    if (!(l->creatures==NULL))
        free_clist(l->creatures);
    for(i=0;i<MAX_X;++i)
        for(j=0;j<MAX_Y;++j)
            {
            if (!(l->space[i][j].creatures==NULL))
                free_creatures(l->space[i][j].creatures);
            if (!(l->space[i][j].items==NULL))
                free_items(l->space[i][j].items);
            if (!(l->space[i][j].special==NULL))
                free_special(l->space[i][j].special);
            }
    if (!(master_llist == NULL))
        {
        if (master_llist->current == l)
            {
            this_llist = master_llist;
            master_llist = master_llist->next;
            free(this_llist);
            }
        else
            for(traverse = previous = master_llist; (!(traverse == NULL));
              traverse = traverse->next)
        if (traverse->current != NULL)
        {
                if (traverse->current == l)
                    {
                    previous->next = traverse->next;
                    free(traverse);
                    traverse = previous;
                    }
                else
                    previous = traverse;
        }
        }
    l->me = NULL;
    free(l);
    }

/* Free all of the linked list, given its head.  */

void free_read_file_text_struct(struct read_file_text_struct *r)
    {
    if (r == NULL || r->me != r)
        return;
    if (r != NULL)
        {
        if (r->next != NULL)
            free_read_file_text_struct(r->next);
        r->me = NULL;
        free(r);
        }
    }

/* Free the pointer structure used to regenerate data from a save file, but */
/* not the new data themselves.  */

static void free_pointer_struct(struct pointer_struct *pointer)
    {
    if (pointer == NULL || pointer->me != pointer)
        return;
    if (!(pointer->next == NULL))
        free_pointer_struct(pointer->next);
    pointer->me = NULL;
    free(pointer);
    }

void free_string_array(char **string_array, long number_of_items)
    {
    int i;
    for(i = 0; i < number_of_items; ++i)
        free(string_array[i]);
    free(string_array);
    }

/* Free a tile used in the dungeon display */

void free_tile(struct tile *the_tile)
	{
	if (the_tile != NULL)
		{
		if (the_tile->me == the_tile)
			{
			if (the_tile->next != NULL)
				free_tile(the_tile->next);
			free(the_tile);
			}
		else
			return;
		}
	}

