/***************************************************************************/
/* The Minstrel's Song is (c) 1998-2005 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.                 */
/***************************************************************************/

#include "tms.h"
#ifndef MAC_68k
#include <sys/types.h>
#endif
#ifdef TMS_CURSES
#include <curses.h>
#endif
#include <stdio.h>
#include <math.h>
#include <string.h>

#ifdef CODE_SEGMENT
#pragma segment DISPLAY_C
#endif

#ifdef GTK
struct tile **dungeon_image;
long *dungeon_image_changed;
#endif

char tgetc(void)
    {
    game_variables->messages_printed = 0;
    game_variables->fs_messages_printed = 0;
#ifndef GTK
    /* The next line is intended as a possible means of addressing a freezing */
    /* bug.  */
    fflush(stdin);
#ifdef TMS_CURSES
    return getch();
#endif
#ifdef MAC_68k
    while (menu_input == NULL)
        {
        Delay(1, 0);
        event_check();
        }
    return menu_pop_char();
#endif
#endif
    return 0;
    }

void ensure_synced_screen(void)
        {
#ifdef SPECIAL_WNOUTREFRESH
        wrefresh(top_text);
#endif
        }

long pmsg_internal(char *message, long flags)
    {
#ifdef GTK
    long flags_target;
    flags_target = flags;
    new_pmsg(message);
    return 1;
#else
    long i, success;
    char local_message[256];
#ifdef DEBUG
    check_string(message);
#endif
    ensure_synced_screen();
    strcpy(local_message, message);
    /* Strip a trailing newline */
    for(i=0;i<256;++i)
	if ((local_message[i] == 10) || (local_message[i] == 13) ||
	  (local_message[i] == 0))
	    {
	    local_message[i] = 0;
	    i = 256;
	    }
    /* Capitalize the first letter, if appropriate.  */
    if ((flag_check(flags, PMSG_CAPITALIZE)) && (local_message[0] >= 'a') &&
      (local_message[0] <= 'z'))
	local_message[0] += 'A' - 'a';
    if (flag_check(flags, PMSG_READ_FILE))
	success = READ_FILE_CONTINUE;
    else
	success = TRUE;
    if (flag_check(flags, PMSG_FULLSCREEN))
	{
#ifdef GTK
	window_puts(WINDOW_STDSCR, game_variables->fs_messages_printed,
	  local_message);
#else
    mvprintw(min(window_lines() - 1, game_variables->fs_messages_printed - 1), 0,
      "%s\n", local_message);
    clrtoeol();
#endif
	++game_variables->fs_messages_printed;
#ifdef GTK
	if (game_variables->fs_messages_printed >= window_lines() - 2)
#else
	if (game_variables->fs_messages_printed >= window_lines() - 2)
#endif
	    {
	    if ((flag_check(game_variables->flags, MESSAGES_PROMPT)) &&
	      (flag_check(flags, PMSG_PROMPT)))
		{
		game_variables->fs_messages_printed=0;
		if (flag_check(flags, PMSG_READ_FILE))
		    success=rf_fs_morewait(); /* have morewait refresh */
		else
		    success=fs_morewait(); /* have morewait refresh */
#ifdef GTK
		window_clear();
#else
        window_clear();
#endif
		}
	    else
		{
#ifdef GTK
		window_scroll(WINDOW_STDSCR);
#endif
		--game_variables->fs_messages_printed;
		}
	    }
	}
    else
	{
	++game_variables->messages_printed;
#ifdef GTK
	window_scroll(WINDOW_TOP_TEXT);
	window_puts(WINDOW_TOP_TEXT, 4, local_message);
#else
    mvwprintw(top_text, 5, 0, "%s\n", local_message);
    wclrtoeol(top_text);
#endif
	if (game_variables->messages_printed >
#ifdef GTK
	  window_boundary(WINDOW_TOP_TEXT, WINDOW_MAX_Y))
#else
      5)
#endif
	    {
	    if ((flag_check(game_variables->flags, MESSAGES_PROMPT)) &&
	      (flag_check(flags, PMSG_PROMPT)))
		{
		game_variables->fs_messages_printed=0;
		success = morewait(); /* have morewait refresh */
		}
	    }
#ifdef GTK
	window_move_cursor(WINDOW_TOP_TEXT, 0, 5);
#endif
	}
#ifdef GTK
    window_refresh();
#else
    window_refresh();
#endif
    ensure_synced_screen();
    return success;
#endif
    }

long fs_pmsg(char *message)
    {
    return pmsg_internal(message, PMSG_FULLSCREEN|PMSG_PROMPT);
    }

long rf_fs_pmsg(char *message)
    {
    return pmsg_internal(message, PMSG_FULLSCREEN|PMSG_PROMPT|PMSG_READ_FILE);
    }

long cap_fs_pmsg(char *message)
    {
    return pmsg_internal(message, PMSG_CAPITALIZE|PMSG_FULLSCREEN|PMSG_PROMPT);
    }

long pmsg(char *message)
    {
    return pmsg_internal(message, PMSG_CAPITALIZE|PMSG_PROMPT);
    }

long noprompt_pmsg(char *message)
    {
    return pmsg_internal(message, PMSG_CAPITALIZE);
    }

long nocap_pmsg(char *message)
    {
    return pmsg_internal(message, PMSG_PROMPT);
    }

void left_panel_update(void)
    {
    char message[256];
    sprintf(message, "Ag: %ld/%ld",
      player->extended_creature->attributes[AG], player->extended_creature->perm_attributes[AG]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 0, message);
#else
    mvwprintw(left_panel, 0, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "AL: %ld/%ld",
      player->extended_creature->attributes[AL], player->extended_creature->perm_attributes[AL]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 1, message);
#else
    mvwprintw(left_panel, 1, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Ch: %ld/%ld",
      player->extended_creature->attributes[CH], player->extended_creature->perm_attributes[CH]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 2, message);
#else
    mvwprintw(left_panel, 2, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Co: %ld/%ld",
      player->extended_creature->attributes[CO], player->extended_creature->perm_attributes[CO]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 3, message);
#else
    mvwprintw(left_panel, 3, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "De: %ld/%ld",
      player->extended_creature->attributes[DE], player->extended_creature->perm_attributes[DE]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 4, message);
#else
    mvwprintw(left_panel, 4, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "In: %ld/%ld",
      player->extended_creature->attributes[IN], player->extended_creature->perm_attributes[IN]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 5, message);
#else
    mvwprintw(left_panel, 5, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Kn: %ld/%ld",
      player->extended_creature->attributes[KN], player->extended_creature->perm_attributes[KN]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 6, message);
#else
    mvwprintw(left_panel, 6, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Me: %ld/%ld",
      player->extended_creature->attributes[ME], player->extended_creature->perm_attributes[ME]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 7, message);
#else
    mvwprintw(left_panel, 7, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Pe: %ld/%ld",
      player->extended_creature->attributes[PE], player->extended_creature->perm_attributes[PE]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 8, message);
#else
    mvwprintw(left_panel, 8, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Sp: %ld/%ld",
      player->extended_creature->attributes[SPEED],
      player->extended_creature->perm_attributes[SPEED]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 9, message);
#else
    mvwprintw(left_panel, 9, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "St: %ld/%ld",
      player->extended_creature->attributes[ST], player->extended_creature->perm_attributes[ST]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 10, message);
#else
    mvwprintw(left_panel, 10, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Wi: %ld/%ld",
      player->extended_creature->attributes[WI], player->extended_creature->perm_attributes[WI]);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 11, message);
#else
    mvwprintw(left_panel, 11, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Att: %ld", player->att);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 13, message);
#else
    mvwprintw(left_panel, 13, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Dmg: %ld", player->dmg);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 14, message);
#else
    mvwprintw(left_panel, 14, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Def: %ld", player->def);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 15, message);
#else
    mvwprintw(left_panel, 15, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "HV: %ld/%ld",
      player->extended_creature->phv-player->extended_creature->stati->time[STATUS_WOUNDED],
      player->extended_creature->phv);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 16, message);
#else
    mvwprintw(left_panel, 16, 0, message);
    wclrtoeol(left_panel);
#endif
    sprintf(message, "Wgt: %ld", player->extended_creature->curr_weight/LB);
#ifdef GTK
    window_puts(WINDOW_LEFT_PANEL, 17, message);
#else
    mvwprintw(left_panel, 17, 0, message);
    wclrtoeol(left_panel);
#endif
    }

#ifdef GTK
GtkWidget *bottom_panel(void)
#else
void bottom_panel_update(void)
#endif
    {
    char day[256], month[256], season[256], X, stati[256],
        mounted[256], minutes[256], result[256], temp[256], space_desc[256],
        room_desc[256], level_desc[256], mount_desc[256];
    long hours, monthday, year, elapsed;
#ifdef GTK
    GtkWidget *current_label, *result_widget;
    result_widget = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(result_widget);
#endif
    stati[0] = '\0';
    mounted[0] = '\0';
    minutes[0] = '\0';
    elapsed=game_variables->elapsed_units;
    /* The room string may sometimes be empty.  Here we check if it is empty, */
    /* and print the appropriate string.  */
    find_string(current_level->environment_name, level_desc);
    find_string(current_level->space[player->x][player->y].space_description,
      space_desc);
    find_string(current_level->space[player->x][player->y].room_description,
      room_desc);
    /* Capitalize first letter, if need be.  */
    if ((space_desc[0] >= 'a') && (space_desc[0] <= 'z'))
        space_desc[0] += 'A' - 'a';
    if (!(room_desc[0] == 0))
#ifdef HAS_SNPRINTF
        snprintf(result, 255, "%s in %s, in %s", space_desc, room_desc, level_desc);
#else
        sprintf(result, "%s in %s, in %s", space_desc, room_desc, level_desc);
#endif
    else
#ifdef HAS_SNPRINTF
        snprintf(result, 255, "%s, in %s", space_desc, level_desc);
#else
        sprintf(result, "%s, in %s", space_desc, level_desc);
#endif
    result[255] = '\0';
    temp[0] = '\0';
    if (flag_check(current_level->flags, PRINT_LEVEL_NUMBER))
        {
        if (current_level->power_level > 0)
#ifdef HAS_SNPRINTF
            snprintf(temp, 255, " (level %ld)", current_level->power_level);
#else
            sprintf(temp, " (level %ld)", current_level->power_level);
#endif
        else if (current_level->power_level < 0)
#ifdef HAS_SNPRINTF
            snprintf(temp, 255, " (floor %ld)", (0 - current_level->power_level));
#else
            sprintf(temp, " (floor %ld)", (0 - current_level->power_level));
#endif
        else
#ifdef HAS_SNPRINTF
            snprintf(temp, 255, " (ground floor)");
#else
            sprintf(temp, " (ground floor)");
#endif
        temp[255] = '\0';
        }
    strncat(result, temp, max(0, 255 - strlen(result)));
    result[255] = '\0';
    strncat(result, ".", max(0, 255 - strlen(result)));
    result[255] = '\0';
#ifdef GTK
    current_label = gtk_label_new(result);
    g_object_set(G_OBJECT(current_label), "xalign", 0.0f, NULL);
    gtk_widget_show(current_label);
    gtk_box_pack_start(GTK_BOX(result_widget), current_label, FALSE, FALSE, 0);
#else
#ifdef MAC_68k
    window_puts(WINDOW_BOTTOM_PANEL, 0, result);
#else
    mvwprintw(bottom_panel, 0, 0, result);
    wclrtoeol(bottom_panel);
#endif
#endif
    /* My modular arithmetic may be buggy */
#ifdef HAS_SNPRINTF
    snprintf(minutes, 255, "%ld%ld",
      ((((elapsed/(UNITS_PER_SECOND*60L)) % 60L)) / 10L),
      (((elapsed/(UNITS_PER_SECOND*60L)) % 60L) -
      (10L * (((elapsed/(UNITS_PER_SECOND*60L)) % 60L) / 10L))));
#else
    sprintf(minutes, "%ld%ld",
      ((((elapsed/(UNITS_PER_SECOND*60L)) % 60L)) / 10L),
      (((elapsed/(UNITS_PER_SECOND*60L)) % 60L) -
      (10L * (((elapsed/(UNITS_PER_SECOND*60L)) % 60L) / 10L))));
#endif
    minutes[255] = '\0';
    /* (60 + (((elapsed/(UNITS_PER_SECOND*60L)) % 60L)) / 10L),
    (60 + (((elapsed/(UNITS_PER_SECOND*60L)) % 60L)) -
    (10L * (10 + (((elapsed/(UNITS_PER_SECOND*60L)) % 60L) / 10L))))); */
    hours=((elapsed/(UNITS_PER_SECOND*3600L))%12);
    if (hours == 0)
        hours = 12;
    monthday=((game_variables->elapsed_days)%28);
    if (monthday == 0)
        monthday = 28;
    year=((game_variables->elapsed_days)/YEAR);
    if ((elapsed/(UNITS_PER_SECOND*3600L))%24 ==
    (elapsed/(UNITS_PER_SECOND*3600))%12) X = 'A'; else X = 'P';
    switch ((game_variables->elapsed_days)%7)
        {
        case 0: strncpy(day,"Sunday", 255); break;
        case 1: strncpy(day,"Monday", 255); break;
        case 2: strncpy(day,"Tuesday", 255); break;
        case 3: strncpy(day,"Wednesday", 255); break;
        case 4: strncpy(day,"Thursday", 255); break;
        case 5: strncpy(day,"Friday", 255); break;
        case 6: strncpy(day,"Saturday", 255); break;
        }
    day[255] = '\0';
    switch ((game_variables->elapsed_days/MONTH)%12)
        {
        case 0: strncpy(month,"January", 255); break;
        case 1: strncpy(month,"February", 255); break;
        case 2: strncpy(month,"March", 255); break;
        case 3: strncpy(month,"April", 255); break;
        case 4: strncpy(month,"May", 255); break;
        case 5: strncpy(month,"June", 255); break;
        case 6: strncpy(month,"July", 255); break;
        case 7: strncpy(month,"August", 255); break;
        case 8: strncpy(month,"September", 255); break;
        case 9: strncpy(month,"October", 255); break;
        case 10: strncpy(month,"November", 255); break;
        case 11: strncpy(month,"December", 255); break;
        }
    month[255] = '\0';
    switch (((game_variables->elapsed_days - MONTH)/84L)%4)
        {
        case 0: strncpy(season,"summer", 255); break;
        case 1: strncpy(season,"autumn", 255); break;
        case 2: strncpy(season,"winter", 255); break;
        case 3: strncpy(season,"spring", 255); break;
        }
    season[255] = '\0';
    sprintf(result, "%ld:%s %cM, %s, %s %ld, %s, year %ld",
      hours, minutes, X, day, month, monthday, season, year);
#ifdef GTK
    current_label = gtk_label_new(result);
    g_object_set(G_OBJECT(current_label), "xalign", 0.0f, NULL);
    gtk_widget_show(current_label);
    gtk_box_pack_start(GTK_BOX(result_widget), current_label, FALSE, FALSE, 0);
#else
#ifdef MAC_68k
    window_puts(WINDOW_BOTTOM_PANEL, 1, result);
#else
    mvwprintw(bottom_panel, 1, 0, result);
    wclrtoeol(bottom_panel);
#endif
#endif
    /* Deal with stati here; I haven't put the descriptive code in yet. */
    /* Separate different stati by two spaces.  */
    stati[0] = '\0';
    if (player->extended_creature->mount)
        {
    if (!(player->extended_creature->mount->name == NULL))
        {
#ifdef HAS_SNPRINTF
        snprintf(mounted, 255, "riding %s  ",
          player->extended_creature->mount->name->text);
#else
        sprintf(mounted, "riding %s  ",
          player->extended_creature->mount->name->text);
#endif
        mounted[255] = '\0';
        }
    else
        {
        if (!(player->extended_creature->mount->extended_creature == NULL))
        strncpy(mount_desc, player->extended_creature->mount->extended_creature->species, 255);
        else
        strncpy(mount_desc,
          master_creatures[player->extended_creature->mount->identity]->species, 255);
        mount_desc[255] = '\0';
            if ((mount_desc[0] == 'a') ||
            (mount_desc[0] == 'e') ||
            (mount_desc[0] == 'i') ||
            (mount_desc[0] == 'o') ||
            (mount_desc[0] == 'u') ||
            (mount_desc[0] == 'A') ||
            (mount_desc[0] == 'E') ||
            (mount_desc[0] == 'I') ||
            (mount_desc[0] == 'O') ||
            (mount_desc[0] == 'U'))
#ifdef HAS_SNPRINTF
                snprintf(mounted, 255, "riding an %s  ", mount_desc);
#else
                sprintf(mounted, "riding an %s  ", mount_desc);
#endif
            else
#ifdef HAS_SNPRINTF
                snprintf(mounted, 255, "riding a %s  ", mount_desc);
#else
                sprintf(mounted, "riding a %s  ", mount_desc);
#endif
            mounted[255] = '\0';
            }
        }
    if (flag_check(player->flags, ABOVE_GROUND))
        {
        strncat(stati, "flying   ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_HUNGRY) > 0)
        switch(status_check(player, STATUS_HUNGRY))
            {
            case 1:
                strncat(stati, "hungry  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 2:
                strncat(stati, "very hungry  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 3:
                strncat(stati, "ravenous  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 4:
                strncat(stati, "weak  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            default:
                strncat(stati, "starving  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            }
    if (status_check(player, STATUS_THIRSTY) > 0)
        switch(status_check(player, STATUS_THIRSTY))
            {
            case 1:
                strncat(stati, "thirsty  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 2:
                strncat(stati, "very thirsty  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 3:
                strncat(stati, "dehydrated  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 4:
                strncat(stati, "severely dehydrated  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            default:
                strncat(stati, "dying of thirst  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            }
    if (flag_check(player->flags, SLEEPING))
        {
        strncat(stati, "asleep  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    else
        if (status_check(player, STATUS_TIRED) > 0)
            switch(status_check(player, STATUS_TIRED))
                {
                case 1:
                    strncat(stati, "drowsy  ", max(0, 255 - strlen(stati)));
                    stati[255] = '\0';
                break;
                case 2:
                    strncat(stati, "tired  ", max(0, 255 - strlen(stati)));
                    stati[255] = '\0';
                break;
                case 3:
                    strncat(stati, "very sleepy  ", max(0, 255 - strlen(stati)));
                    stati[255] = '\0';
                break;
                case 4:
                    strncat(stati, "barely awake  ", max(0, 255 - strlen(stati)));
                    stati[255] = '\0';
                break;
                default:
                    strncat(stati, "dangerously sleepy  ", max(0, 255 - strlen(stati)));
                    stati[255] = '\0';
                break;
                }
    if (status_check(player, STATUS_FATIGUED) > 0)
        switch(status_check(player, STATUS_FATIGUED))
            {
            case 1:
                strncat(stati, "fatigued  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 2:
                strncat(stati, "worn out  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 3:
                strncat(stati, "exhausted  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 4:
                strncat(stati, "totally exhausted  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            default:
                strncat(stati, "dangerously fatigued  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            }
    if (status_check(player, STATUS_COLD) > 0)
        switch(status_check(player, STATUS_COLD))
            {
            case 1:
                strncat(stati, "chilly  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 2:
                strncat(stati, "cold  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 3:
                strncat(stati, "chilled to the bone  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 4:
                strncat(stati, "hypothermic  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            default:
                strncat(stati, "dying of hypothermia  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            }
    if (status_check(player, STATUS_IMMOBILE) > 0)
        {
        strncat(stati, "immobile  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_LOST) > 0)
        {
        strncat(stati, "lost  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_PASSED_OUT) > 0)
        {
        strncat(stati, "passed out  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_STINKY) > 0)
        {
        strncat(stati, "stinky  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_ALCOHOL) > 0)
        switch(status_check(player, STATUS_ALCOHOL))
            {
            case 1:
                strncat(stati, "relaxed  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 2:
                strncat(stati, "merry  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 3:
                strncat(stati, "a bit too merry  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 4:
                strncat(stati, "rather tipsy  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
            break;
            case 5:
                strncat(stati, "drunk  ", max(0, 255 - strlen(stati)));
                stati[255] = '\0';
                flag_set(&player->flags, HAS_SINNED);
            break;
            }
    if (status_check(player, STATUS_ADRENALINE) > 0)
        {
        strncat(stati, "on adrenaline surge  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_BREATHE_WATER) > 0)
        {
        strncat(stati, "breathing water  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_COFFEE) > 0)
        {
        strncat(stati, "on caffeine  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_COFISA) > 0)
        {
        strncat(stati, "on cofisa tea  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_HELSRIANA) > 0)
        {
        strncat(stati, "on helsriana tea  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_HUMMER) > 0)
        {
        strncat(stati, "repelling animals  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    if (status_check(player, STATUS_TALINOR) > 0)
        {
        strncat(stati, "on talinor tea  ", max(0, 255 - strlen(stati)));
        stati[255] = '\0';
        }
    sprintf(result, "%s%s", mounted, stati);
#ifdef GTK
    current_label = gtk_label_new(result);
    g_object_set(G_OBJECT(current_label), "xalign", 0.0f, NULL);
    gtk_widget_show(current_label);
    gtk_box_pack_start(GTK_BOX(result_widget), current_label, FALSE, FALSE, 0);
    return result_widget;
#else
#ifdef MAC_68k
    window_puts(WINDOW_BOTTOM_PANEL, 2, result);
#else
    mvwprintw(bottom_panel, 2, 0, result);
    wclrtoeol(bottom_panel);
#ifdef NO_WNOUTREFRESH
    wrefresh(bottom_panel);
#else
    wnoutrefresh(bottom_panel);
#endif
#endif
#endif
    }

#ifndef GTK
long image_character(long x, long y)
    {
    long dispchar = ' ';
    long result, saw_creature;
    saw_creature = FALSE;
    if ((x < 0) || (x >= MAX_X) || (y < 0) || (y >= MAX_Y))
        return ' ';
    if (flag_check(current_level->space[x][y].flags, KNOWN))
        dispchar= current_level->space[x][y].image;
    if (can_see(player, x, y))
        {
        flag_set(&current_level->space[x][y].flags, KNOWN);
        if (current_level->space[x][y].creatures)
            {
            if (flag_check(current_level->space[x][y].creatures->current->
              flags, HIDDEN))
                {
                result = skill_check(player, player->extended_creature->skills->
                  av[SEARCH], DIFFICULT,
                  0);
                if (result == 1)
                    {
                    dispchar=C_CANT_TELL;
                    saw_creature = TRUE;
                    }
                else if (result > 1)
                    {
                    dispchar = master_creatures[current_level->space[x][y].
                      creatures->current->identity]->image;
                    saw_creature = TRUE;
                    flag_reset(&current_level->space[x][y].creatures->current->
                      flags, HIDDEN);
                    }
                }
            else
                {
                dispchar = master_creatures[current_level->space[x][y].
                  creatures->current->identity]->image;
                saw_creature = TRUE;
                }
            }
        if (!(saw_creature))
            {
            if ((current_level->space[x][y].items)
              && (current_level != countryside))
                {
                if
                (current_level->space[x][y].items->next)
                    dispchar=C_PILE;
                else
                    dispchar=orig_items[current_level->space[x][y].
                      items->current->identity]->image;
                }
            else
                {
                if (((current_level->space[x][y].au) ||
                  (current_level->space[x][y].el) ||
                  (current_level->space[x][y].si) ||
                  (current_level->space[x][y].cu))
                  && (current_level != countryside))
                    dispchar=C_MONEY;
                else
                    {
                    if ((flag_check(current_level->space[x][y].
                      flags, S_CLOAKED_BY_MIST)) &&
                      (current_level->space[x][y].image==C_FLOOR))
                        dispchar=C_MIST;
                    else
                        dispchar=current_level->space[x][y].image;
                    }
                }
            }
        }
    else if (can_see_badly(player, x, y))
        {
        flag_set(&current_level->space[x][y].flags, KNOWN);
        if (current_level->space[x][y].creatures)
            {
            if (flag_check(current_level->space[x][y].creatures->current->
              flags, HIDDEN))
                {
                if (skill_check(player, player->extended_creature->skills->av[SEARCH],
                  VERY_DIFFICULT,
                  0))
                    dispchar=C_CANT_TELL;
                }
            else
                dispchar=C_CANT_TELL;
            }
        if (dispchar != C_CANT_TELL)
            {
            if ((current_level->space[x][y].items)
              && (current_level != countryside))
                dispchar=C_CANT_TELL;
            else
                {
                if ((flag_check(current_level->space[x][y].flags,
                  S_CLOAKED_BY_MIST)) &&
                  (current_level->space[x][y].image==C_FLOOR))
                    dispchar=C_MIST;
                else
                    dispchar=current_level->space[x][y].image;
                }
            }
        }
    else if (current_level->space[x][y].creatures != NULL)
        {
        result = skill_check(player, player->extended_creature->skills->av[HEAR_NOISES],
          distance(player->x, player->y, x, y) * current_level->scale +
          ONE_NOTCH * (!(flag_check(current_level->space[x][y].creatures->
          current->flags, INACTIVE))) + ONE_NOTCH *
          (!(clear_path(player->x, player->y, x, y))) + ONE_NOTCH *
          (flag_check(current_level->space[x][y].creatures->current->flags,
          HIDDEN)), 0);
        if (result == 1)
            {
            dispchar = 'X';
            saw_creature = TRUE;
            }
        else if (result > 1)
            {
            dispchar = master_creatures[current_level->space[x][y].creatures->
              current->identity]->image;
            saw_creature = TRUE;
            }
        }

/*
 *
 * Player can always "see" himself, even if it's totally dark, he's in the
 * countryside, etc.
 *
 */

    if ((player->x == x) && (player->y == y))
        {
        dispchar=player->extended_creature->image;
        flag_set(&current_level->space[x][y].flags, KNOWN);
        }
    if ((flag_check(current_level->space[x][y].flags, KNOWN)) &&
      (flag_check(game_variables->flags, ALWAYS_DISPLAY_STAIRCASES)) &&
      ((current_level->space[x][y].image == C_STAIRWELL_UP) ||
      (current_level->space[x][y].image == C_STAIRWELL_DOWN)))
        dispchar = current_level->space[x][y].image;
    if (saw_creature)
        {
#ifdef MAC_68k
        return dispchar;
#else
        return dispchar | A_BOLD;
#endif
        }
    else
        {
#ifdef MAC_68k
        return dispchar;
#else
        return dispchar | A_DIM;
#endif
        }
    }
#endif

struct tile *find_tile(long x, long y)
	{
	struct tile *result;
	struct clist *cltraverse;
	struct ilist *iltraverse;
	long halo_id, skill_check_result;
	result = NULL;
	halo_id = NONE;
	if ((0 <= x) && (x < MAX_X) && (0 <= y) && (y < MAX_Y))
		{
		if ((flag_check(current_level->space[x][y].flags, KNOWN))
		  || (can_see(player, x, y)) || (can_see_badly(player, x, y)))
#ifdef GTK
			prepend_icon(&result, current_level->space[x][y].icon);
#else
			append_icon(&result, current_level->space[x][y].icon);
#endif
		if ((flag_check(current_level->space[x][y].flags, KNOWN))
		  && (!(can_see(player, x, y))) && (!(can_see_badly(player, x, y))) &&
		  (!(flag_check(current_level->space[x][y].flags, CURR_ILLUMINATED))))
#ifdef GTK
		  	prepend_icon(&result, ICON_DARKNESS_MASK);
#else
		  	append_icon(&result, ICON_DARKNESS_MASK);
#endif
		if ((can_see(player, x, y)) || (can_see_badly(player, x, y)))
			{
			flag_set(&current_level->space[x][y].flags, KNOWN);
			if (can_see(player, x, y))
				{
				if (current_level->space[x][y].cu > 0)
					{
#ifdef GTK
					prepend_icon(&result, ICON_COPPER_COINS);
#else
					append_icon(&result, ICON_COPPER_COINS);
#endif
					halo_id = ICON_HALO_MONEY;
					}
				if (current_level->space[x][y].si > 0)
					{
#ifdef GTK
					prepend_icon(&result, ICON_SILVER_COINS);
#else
					append_icon(&result, ICON_SILVER_COINS);
#endif
					halo_id = ICON_HALO_MONEY;
					}
				if (current_level->space[x][y].el > 0)
					{
#ifdef GTK
					prepend_icon(&result, ICON_ELECTRUM_COINS);
#else
					append_icon(&result, ICON_ELECTRUM_COINS);
#endif
					halo_id = ICON_HALO_MONEY;
					}
				if (current_level->space[x][y].au > 0)
					{
#ifdef GTK
					prepend_icon(&result, ICON_GOLD_COINS);
#else
					append_icon(&result, ICON_GOLD_COINS);
#endif
					halo_id = ICON_HALO_MONEY;
					}
				}
			if (current_level->space[x][y].items)
				{
				halo_id = ICON_HALO_ITEM;
				for(iltraverse = current_level->space[x][y].items;
				  iltraverse != NULL; iltraverse = iltraverse->next)
#ifdef GTK
				  	prepend_icon(&result, 
				  	  orig_items[iltraverse->current->identity]->icon);
#else
				  	append_icon(&result, 
				  	  orig_items[iltraverse->current->identity]->icon);
#endif
				}
			if (current_level->space[x][y].creatures)
			    {
			    for(cltraverse = current_level->space[x][y].creatures;
			      cltraverse != NULL; cltraverse = cltraverse->next)
			      	{
				    if (flag_check(cltraverse->current->flags, HIDDEN))
						{
						skill_check_result = skill_check(player, player->extended_creature->skills->
						  av[SEARCH], DIFFICULT,
						  0);
						if (skill_check_result >= 1)
						    {
#ifdef GTK
						    prepend_icon(&result, find_icon(cltraverse->current));
#else
						    append_icon(&result, find_icon(cltraverse->current));
#endif
						    if (flag_check(cltraverse->current->flags, HUMAN))
						      	halo_id = ICON_HALO_PERSON;
						    else if (halo_id != ICON_HALO_PERSON)
					    		halo_id = ICON_HALO_ANIMAL;
						    }
						}  
			 	    else
					    {
#ifdef GTK
				 		prepend_icon(&result, find_icon(cltraverse->current));
#else
				 		append_icon(&result, find_icon(cltraverse->current));
#endif
					    if (flag_check(cltraverse->current->flags, HUMAN))
					      	halo_id = ICON_HALO_PERSON;
					    else if (halo_id != ICON_HALO_PERSON)
					    	halo_id = ICON_HALO_ANIMAL;
					    }
				    }
				}
			if (flag_check(current_level->space[x][y].flags, S_CLOAKED_BY_MIST))
				{
				if (!(flag_check(current_level->space[x][y].flags, CURR_ILLUMINATED)))
#ifdef GTK
					prepend_icon(&result, ICON_FOG_AND_DARKNESS_MASK);
#else
					append_icon(&result, ICON_FOG_AND_DARKNESS_MASK);
#endif
				else
#ifdef GTK
					prepend_icon(&result, ICON_FOG_MASK);
#else
					append_icon(&result, ICON_FOG_MASK);
#endif
				}
			else if (!(flag_check(current_level->space[x][y].flags, CURR_ILLUMINATED)))
				{
#ifdef GTK
				prepend_icon(&result, ICON_DARKNESS_MASK);
#else
				append_icon(&result, ICON_DARKNESS_MASK);
#endif
				}
			/* Work in halo for features.  */
			}
 		else if (current_level->space[x][y].creatures != NULL)
			{
			for(cltraverse = current_level->space[x][y].creatures; cltraverse != NULL;
			  cltraverse = cltraverse->next)
			  	{
				skill_check_result = skill_check(player, player->extended_creature->skills->av[HEAR_NOISES],
				  distance(player->x, player->y, x, y) * current_level->scale +
				  ONE_NOTCH * (!(flag_check(cltraverse->current->flags, INACTIVE))) +
				  ONE_NOTCH * (!(clear_path(player->x, player->y, x, y))) + ONE_NOTCH *
				  (flag_check(cltraverse->current->flags, HIDDEN)), 0);
				if (skill_check_result == 1)
#ifdef GTK
					prepend_icon(&result, ICON_CANT_TELL);
#else
					append_icon(&result, ICON_CANT_TELL);
#endif
				else if (skill_check_result > 1)
#ifdef GTK
					prepend_icon(&result, find_icon(cltraverse->current));
#else
					append_icon(&result, find_icon(cltraverse->current));
#endif
				}
			}
/*
 *
 * Player can always "see" himself, even if it's totally dark, he's in the
 * countryside, etc.
 *
 */
		if ((player->x == x) && (player->y == y))
			{
#ifdef GTK
			prepend_icon(&result, player->extended_creature->icon);
#else
			append_icon(&result, player->extended_creature->icon);
#endif
			halo_id = ICON_HALO_PLAYER;
			}
		if (halo_id != NONE)
#ifdef GTK
			prepend_icon(&result, halo_id);
#else
			append_icon(&result, halo_id);
#endif
        if (cursor->should_display && (x == cursor->x && y == cursor->y))
#ifdef GTK
			prepend_icon(&result, ICON_CURSOR_SELECT);
#else
			append_icon(&result, ICON_CURSOR_SELECT);
#endif
		}
	if (result == NULL)
#ifdef GTK
		prepend_icon(&result, ICON_UNKNOWN);
#else
		append_icon(&result, ICON_UNKNOWN);
#endif
	return result;
	}

#ifdef GTK
GtkWidget *dungeon_map(long x, long y, long should_use_event_boxes)
    {
    /*
    GtkWidget *container;
    */
    long i, j, k, l;
    struct tile *curr_tile, *traverse;
    for(i = 0; i < MAX_X; ++i)
        for(j = 0; j < MAX_Y; ++j)
            if (can_see(player, x, y))
                flag_set(&current_level->space[x][y].flags, KNOWN);
    GtkWidget *container, *current_icon, *result;
    result = gtk_table_new(x, y, FALSE);
    if ((flag_check(game_variables->flags, PLAYER_AT_CENTER)) ||
      (abs(player->x - game_variables->display_center_x) >
      (game_variables->old_window_dungeon_columns) / 2 - 2)
      || (abs(player->y - game_variables->display_center_y) > 
      (game_variables->old_window_dungeon_lines)/2 -
      2))
	{
	game_variables->display_center_x=player->x;
	game_variables->display_center_y=player->y;
	}
	game_variables->old_window_dungeon_columns = x;
	game_variables->old_window_dungeon_lines = y;
    for(i=0; i < game_variables->old_window_dungeon_columns;++i)
	for(j=0; j < game_variables->old_window_dungeon_lines;++j)
	    {
	    k=i+game_variables->display_center_x-
	    (window_dungeon_columns()) / 2;
	    l=j+game_variables->display_center_y-
	    (window_dungeon_lines()) / 2;
	    curr_tile = find_tile(k, l);
            for(traverse = curr_tile; traverse != NULL; traverse =
              traverse->next)
                {
                current_icon = lookup_icon(traverse->icon_id);
                if (should_use_event_boxes)
                    {
                    container = gtk_event_box_new();
                    gtk_event_box_set_visible_window(GTK_EVENT_BOX(container),
                      FALSE);
                    gtk_container_add(GTK_CONTAINER(container), current_icon);
                    gtk_widget_show(container);
                    g_signal_connect_swapped(G_OBJECT(container),
                      "button_press_event",
                      G_CALLBACK(dungeon_map_clicked), (void *)
                      (l * MAX_X + k));
                    gtk_table_attach_defaults(GTK_TABLE(result), container,
                      i, i + 1, j, j + 1);
                    }
                else
                    {
                    gtk_table_attach_defaults(GTK_TABLE(result), current_icon,
                      i, i + 1, j, j + 1);
                    }
                gtk_widget_show(current_icon);
                }
	    }
    return result;
    }
#else
void dungeon_display_update(void)
    {
    long dispchar;
    long i, j, k, l;
    if ((flag_check(game_variables->flags, PLAYER_AT_CENTER)) ||
      (abs(player->x - game_variables->display_center_x) > (window_columns() - 18) / 2 - 2)
      || (abs(player->y - game_variables->display_center_y) > (window_lines() - 12)/2 -
      2))
        {
        game_variables->display_center_x=player->x;
        game_variables->display_center_y=player->y;
        }
    for(i=0;i<window_columns()-16;++i)
        for(j=0;j<window_lines()-10;++j)
            {
            dispchar=C_UNKNOWN;
            if (!(((game_variables->display_center_x - (window_columns()-16)/2 + i) < 0) ||
            ((game_variables->display_center_x - (window_columns()-15)/2 + i) > MAX_X-1) ||
            ((game_variables->display_center_y - (window_lines()-10)/2 + j) < 0) ||
            ((game_variables->display_center_y - (window_lines()-9)/2 + j) > MAX_Y-1)))
                {
                k=i+game_variables->display_center_x-(window_columns()-16)/2;
                l=j+game_variables->display_center_y-(window_lines()-10)/2;
                dispchar = image_character(k, l);
                }
            mvwaddch(dungeon, j, i, dispchar);
            }
#ifdef NO_WNOUTREFRESH
    wrefresh(dungeon);
#else
    wnoutrefresh(dungeon);
#endif
    }
#endif

void top_text_update(void)
    {
#ifndef GTK
#ifdef NO_WNOUTREFRESH
    wrefresh(top_text);
#else
    wnoutrefresh(top_text);
#endif
#endif
    }

void murefresh(void)
    {
#ifdef GTK
    window_move_cursor(WINDOW_DUNGEON,
      ((player->x - game_variables->display_center_x) +
      window_boundary(WINDOW_DUNGEON, WINDOW_MAX_X) -
      window_boundary(WINDOW_DUNGEON, WINDOW_MIN_X)),
      ((player->y - game_variables->display_center_y) +
      window_boundary(WINDOW_DUNGEON, WINDOW_MAX_Y) -
      window_boundary(WINDOW_DUNGEON, WINDOW_MIN_Y)));
#else
    move(((player->y - game_variables->display_center_y) + 6 + (window_lines()-10)/2),
      ((player->x - game_variables->display_center_x) + 13 + (window_columns()-16)/2));
#endif
    /* Move the cursor to the player */
    if (flag_check(game_variables->flags, MAX_REFRESH))
#ifdef GTK
        window_refresh();
#else
        window_refresh();
#endif
    }

long morewait(void)
    {
#ifdef GTK
    return 1;
#else
    char c;
#ifdef MAC_68k
    window_puts(WINDOW_TOP_TEXT, 5, "-- More -- (hit space to continue)");
    window_refresh();
#else
    mvwprintw(top_text, 5, 0, "-- More -- (hit space to continue)");
    wclrtoeol(top_text);
    wrefresh(top_text);
#endif
    for(c = tgetc();(!((c == ' ') || (c == 'q') || (c == ESCAPE)));
#ifndef MAC_68k
      putchar(7)
#endif
      )
        {
        c = tgetc();
        fflush(stdout);
        }
#ifdef MAC_68k
    window_puts(WINDOW_TOP_TEXT, 5, "");
#else
    mvwprintw(top_text, 5, 0, "");
    wclrtoeol(top_text);
    wrefresh(top_text);
#endif
    return (!((c == 'q') || (c == 27)));
#endif
    }

long fs_morewait(void)
    {
#ifdef GTK
    return 1;
#else
    char c;
#ifdef MAC_68k
    window_puts(WINDOW_STDSCR, window_boundary(WINDOW_STDSCR, WINDOW_MAX_Y),
      "-- More -- (hit space to continue)");
    window_refresh();
#else
    mvprintw(window_lines()-1, 0, "-- More -- (hit space to continue)");
    clrtoeol();
    window_refresh();
#endif
    for(c=tgetc();(!((c == ' ') || (c == 'q') || (c == ESCAPE)));putchar(7))
        {
        c=tgetc();
        fflush(stdout);
        }
#ifdef GTK
    window_puts(WINDOW_STDSCR, window_boundary(WINDOW_STDSCR, WINDOW_MAX_Y),
      "");
    window_refresh();
#else
    mvprintw(window_lines(), 0, "");
    clrtoeol();
    window_refresh();
#endif
    return (!((c == 'q') || (c == 27)));
#endif
    }

long rf_fs_morewait(void)
    {
    char c;
    long my_continue, result;
#ifdef GTK
    c = '\0';
    my_continue = 0;
    result = 0;
    return 1;
#else
    result = 0;
#ifdef MAC_68k
    window_puts(WINDOW_STDSCR, window_boundary(WINDOW_STDSCR, WINDOW_MAX_Y),
      "-- More -- (hit space to continue, '?' for help)");
    window_refresh();
#else
    mvprintw(window_lines()-1, 0, "-- More -- (hit space to continue, '?' for help)");
    clrtoeol();
    window_refresh();
#endif
    my_continue = TRUE;
    while (my_continue)
        {
        c = tgetc();
        my_continue = FALSE;
        switch(c)
            {
            case '?': case 'h':
                result = READ_FILE_HELP;
            break;
            case ' ': case 'f': case '+':
                result = READ_FILE_PAGE_DOWN;
            break;
            case '-': case 'p': case 'b':
                result = READ_FILE_PAGE_UP;
            break;
            case ESCAPE: case 'q':
                result = READ_FILE_QUIT;
            break;
            case REFRESH: case 12 /* ^L */:
                result = READ_FILE_REFRESH;
            break;
            case '\\':
                result = READ_FILE_SEARCH_BACKWARDS;
            break;
            case '/':
                result = READ_FILE_SEARCH_FORWARDS;
            break;
            default:
                putchar(7); /* Bell */
                my_continue = TRUE;
            break;
            }
        }
#ifdef MAC_68k
    window_puts(WINDOW_STDSCR, window_boundary(WINDOW_STDSCR, WINDOW_MAX_Y),
      "");
    window_refresh();
#else
    mvprintw(window_lines()-1, 0, "");
    clrtoeol();
    window_refresh();
#endif
    return result;
#endif
    }

void fs_anykey(void)
    {
#ifdef GTK
    /* For now do nothing from gtk. */
#else
#ifdef MAC_68k
    window_puts(WINDOW_STDSCR, window_boundary(WINDOW_STDSCR, WINDOW_MAX_Y),
      "Hit any key to continue...");
    window_refresh();
#else
    mvprintw(window_lines() - 1, 0, "Hit any key to continue...");
    clrtoeol();
    window_refresh();
#endif
    tgetc();
#ifdef MAC_68k
    window_puts(WINDOW_STDSCR, window_boundary(WINDOW_STDSCR, WINDOW_MAX_Y),
      "");
#else
    mvprintw(window_lines() - 1, 0, "");
    clrtoeol();
    window_refresh();
#endif
#endif
    }

void update_window(void)
    {
#ifdef GTK
#ifdef UNDEFINED
    if (!(flag_check(game_variables->flags, DURING_INIT)))
        {
        left_panel_update();
        bottom_panel_update();
        top_text_update();
        dungeon_display_update();
        /*
        window_move_cursor(WINDOW_DUNGEON, ((player->x -
          game_variables->display_center_x) + (window_boundary(WINDOW_DUNGEON,
          WINDOW_MAX_X) - window_boundary(WINDOW_DUNGEON, WINDOW_MIN_X)) /2),
          (player->y - game_variables->display_center_y) +
          (window_boundary(WINDOW_DUNGEON, WINDOW_MAX_Y) -
          window_boundary(WINDOW_DUNGEON, WINDOW_MIN_Y)) / 2);
        */
        }
    else
        window_clear();
    window_refresh();
#endif
#else
    left_panel_update();
    bottom_panel_update();
    top_text_update();
    dungeon_display_update();
    move(((player->y - game_variables->display_center_y) + 6 + (window_lines()-10)/2),
    ((player->x - game_variables->display_center_x) + 13 + (window_columns()-16)/2));
    window_refresh();
#endif
    }

long tmvgetstr(long y, long x, char *result) /* Having trouble with what */
                                                                                         /* curses provides.  */
    {
    long my_continue, position, success;
    char input;
#ifdef GTK
    window_move_cursor(WINDOW_STDSCR, 0, y);
    window_refresh();
#else
    move(y, x);
#endif
    my_continue = TRUE;
    position = -1;
    success = 0;
    while (my_continue)
        {
        input = tgetc();
        switch(input)
            {
            case ESCAPE:
                result[0] = 0;
                my_continue = 0;
                success = 0;
            break;
            case 8: case 127:
#ifdef GTK
                window_assign_character(x + position, y, ' ');
                window_move_cursor(WINDOW_STDSCR, x + position, y);
#else
                mvaddch(y, x + position, ' ');
                move(y, x + position);
#endif
                result[position] = 0;
                position = max(position - 1, -1);
            break;
            case 10: case 13:
                my_continue = 0;
                success = 1;
            break;
            default:
                ++position;
#ifdef GTK
                window_assign_character(x + position, y, input);
#else
                mvaddch(y, x + position, input);
#endif
                result[position] = input;
                result[position + 1] = 0;
            break;
            }
#ifndef GTK
        window_refresh();
#endif
        }
#ifdef GTK
    window_move_cursor(WINDOW_STDSCR, x, y);
#else
    move(y, x);
#endif
    return success;
    }

long tgets(char *result)
    {
    pmsg("");
    ensure_synced_screen();
    return tmvgetstr(4, 0, result);
    }

void projectile_path(long start_x, long start_y, long finish_x, long finish_y,
  char image)
    {
    /*
    long i, total, x, y, disp_x, disp_y;
    total = abs(start_x - finish_x) + abs(start_y - finish_y);
    for(i = 0; i < total; ++i)
        {
        x = start_x + finish_x * (i/total);
        y = start_y + finish_y * (i/total);
        if ((x != start_x) || (y != start_y))
            {
#ifdef GTK
            disp_x = x + game_variables->display_center_x -
              (window_boundary(WINDOW_DUNGEON, WINDOW_MAX_X) -
              window_boundary(WINDOW_DUNGEON, WINDOW_MIN_X)) / 2;
            disp_y = y + game_variables->display_center_y -
              (window_boundary(WINDOW_DUNGEON, WINDOW_MAX_Y) -
              window_boundary(WINDOW_DUNGEON, WINDOW_MIN_Y)) / 2;
            if ((0 <= disp_x) && (disp_x <
              (window_boundary(WINDOW_DUNGEON, WINDOW_MAX_X) -
              window_boundary(WINDOW_DUNGEON, WINDOW_MIN_X)))
              && (0 <= disp_y) &&
              (disp_y < (window_boundary(WINDOW_DUNGEON, WINDOW_MAX_Y) -
              window_boundary(WINDOW_DUNGEON, WINDOW_MIN_Y))))
                {
                window_add_character(WINDOW_DUNGEON, disp_x, disp_y, image);
                window_refresh();
                window_add_character(WINDOW_DUNGEON, disp_x, disp_y,
                  image_character(x, y));
                window_refresh();
                }
#else
            disp_x=x+game_variables->display_center_x-(window_columns()-16)/2;
            disp_y=y+game_variables->display_center_y-(window_lines()-10)/2;
            if ((0 <= disp_x) && (disp_x < window_columns() - 16) && (0 <= disp_y) &&
              (disp_y < window_lines() - 10))
                {
                mvwaddch(dungeon, disp_y, disp_x, image);
                window_refresh();
                mvwaddch(dungeon, disp_y, disp_x, image_character(x, y));
                window_refresh();
                }
#endif
            }
        }
    */
    }
