/***************************************************************************/
/* The Minstrel's Song is (c) 1998-2005 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.                 */
/***************************************************************************/

#include <math.h>
#include <string.h>
#include <stdio.h>
#include "tms.h"

#ifdef CODE_SEGMENT
#pragma segment LEVEL1_C
#endif

struct level *make_level(long level_type, long power_level, long flags, long
  other)
    {
    struct level *result;
    struct llist *this_llist;
    if (!(flag_check(game_variables->flags, DURING_INIT)))
        pmsg("One moment, please, as I dig a new level...");
    if (flag_check(flags, LOAD))
        result = load_level(level_type);
    else
        result = dig_level(level_type, power_level, flags, other);
    switch(level_type)
        {
        case LEVEL_CAVERN: case LEVEL_SMALL_CAVERN:
            populate_animal(result, DUNGEON_ENCOUNTER, power_level, 50);
            populate_item(result, DUNGEON_ITEM_NORMAL, power_level, 200);
            populate_money(result, power_level, 100);
            if (!(flag_check(game_variables->quests_completed,
              COMPLETED_DUNGEON)))
                if (power_level >= 25)
                    if (trandom() % 2 == 0)
                    add_item_to_level(result, FIRESTAR);
        break;
        case LEVEL_MANSION:
            populate_animal(result, DUNGEON_ENCOUNTER, power_level, 500);
            populate_item(result, BUILDING_ITEM_NORMAL, power_level, 200);
            populate_money(result, power_level, 200);
            /*
            if (!(flag_check(game_variables->quests_completed,
              EXPLORED_MANSION)))
                if (power_level == 0 - 7)
                    add_item_to_level(result, SPIDER_SILK_ROBE);
            */
        break;
        case LEVEL_MAZE:
            populate_animal(result, DUNGEON_ENCOUNTER, power_level, 200);
            populate_item(result, DUNGEON_ITEM_NORMAL, power_level, 50);
            populate_money(result, power_level * 7, 100);
            if (!(flag_check(game_variables->quests_completed,
              EXPLORED_MISTRELLI)))
                add_item_to_level(result, DIAMOND_CHALICE);
        break;
        case LEVEL_URVANOVESTILLI_CITY:
            if (other == URVANOVESTILLI_CITY_ELIRE)
                encounter_internal(result, DOLPHIN, 2 + (trandom() % 3));
        break;
        }
    result->power_level = power_level;
    this_llist = (struct llist *) calloc(1, sizeof(struct llist));
    this_llist->me = this_llist;
    this_llist->current = result;
    this_llist->next = master_llist;
    master_llist = this_llist;
    return result;
    }

struct level *load_level(long level_type)
    {
    long i, j, exit_border;
    struct level *result;
    struct small_creature *this_creature;
    char this_row[MAX_X+2], filename[256];
    FILE *infile;
    exit_border = FALSE;
    strncpy(filename, LIB_DIR, 255);
    filename[255] = '\0';
    switch(level_type)
        {
        case LEVEL_CATHEDRAL:
            switch(trandom() % 5 + 1)
                {
                case 1:
                    strncat(filename, FILE_CATHEDRAL_1, max(0, 255 - strlen(filename)));
                break;
                case 2:
                    strncat(filename, FILE_CATHEDRAL_2, max(0, 255 - strlen(filename)));
                break;
                case 3:
                    strncat(filename, FILE_CATHEDRAL_3, max(0, 255 - strlen(filename)));
                break;
                case 4:
                    strncat(filename, FILE_CATHEDRAL_4, max(0, 255 - strlen(filename)));
                break;
                case 5:
                    strncat(filename, FILE_CATHEDRAL_5, max(0, 255 - strlen(filename)));
                break;
                }
        break;
        case LEVEL_COUNTRYSIDE:
            strncat(filename, FILE_COUNTRYSIDE, max(0, 255 - strlen(filename)));
        break;
        case LEVEL_CRYPT:
            switch(trandom() % 3 + 1)
                {
                case 1:
                    strncat(filename, FILE_CRYPT_1, max(0, 255 - strlen(filename)));
                break;
                case 2:
                    strncat(filename, FILE_CRYPT_2, max(0, 255 - strlen(filename)));
                break;
                case 3:
                    strncat(filename, FILE_CRYPT_3, max(0, 255 - strlen(filename)));
                }
        break;
        case LEVEL_FINAL_QUEST:
            strncat(filename, FILE_FINAL_QUEST_DUNGEON, max(0, 255 - strlen(filename)));
        break;
        case LEVEL_NORKRIN_ICE_PEAKS_HIDDEN_POOL:
            strncat(filename, FILE_ICE_PEAKS_POOL, max(0, 255 - strlen(filename)));
        break;
        case LEVEL_OCEAN_ISLE:
            strncat(filename, FILE_OCEAN_ISLE, max(0, 255 - strlen(filename)));
        break;
        case LEVEL_TAVERN:
            switch(trandom() % 4 + 1)
                {
                case 1:
                    strncat(filename, FILE_TAVERN_1, max(0, 255 - strlen(filename)));
                break;
                case 2:
                    strncat(filename, FILE_TAVERN_2, max(0, 255 - strlen(filename)));
                break;
                case 3:
                    strncat(filename, FILE_TAVERN_3, max(0, 255 - strlen(filename)));
                break;
                case 4:
                    strncat(filename, FILE_TAVERN_4, max(0, 255 - strlen(filename)));
                break;
                }
        break;
        }
    filename[255] = '\0';
    infile = fopen(filename, "r");
    if (!(infile))
        {
#ifdef MAC_68k
        pmsg("Error: Unsuccessful attempt to read file:");
        pmsg("");
        pmsg(filename);
        pmsg("");
        pmsg("Aborting...");
#else
        perror("Error: Unsuccessful attempt to read file:");
        perror("");
        perror(filename);
        perror("");
        perror("Aborting...");
#endif
        exit(1);
        }
    result = (struct level *) calloc(1, sizeof(struct level));
    result->me = result;
    result->flags = 0;
    result->scale = 1;
    result->timescale = 1;
    result->race = NONE;
    result->level_type = level_type;
    result->establishment = NULL;
    result->creatures = NULL;
    result->light_sources = NULL;
    result->entrance_message = ST_EMPTY;
    result->exit_message = ST_EMPTY;
    for(i=0;i<MAX_X;++i)
        for(j=0;j<MAX_Y;++j)
            prepare_space(&result->space[i][j], S_FLOOR, result->race);
    for(i = 0; i < NUM_LOCATION_TYPES; ++i)
        {
        result->location[i].flags = 0;
        result->location[i].identity = locations[i].identity;
        result->location[i].x = 0;
        result->location[i].y = 0;
        }
    switch(level_type)
        {
        case LEVEL_CATHEDRAL:
            result->environment_name = ST_A_CATHEDRAL;
            result->entrance_message = ST_YOU_ENTER_A_CATHEDRAL;
            result->exit_message = ST_YOU_LEAVE_A_CATHEDRAL;
            result->race = current_level->race;
            exit_border = TRUE;
        break;
        case LEVEL_COUNTRYSIDE:
            countryside = result;
            result->scale = 5280L * 10L; /* Ten miles */
            result->timescale = MINUTE * 7L;
            flag_set(&result->flags, OUTDOORS);
            result->environment_name = ST_THE_COUNTRYSIDE;
            result->exit_message = ST_YOU_LEAVE_THE_COUNTRYSIDE;
            result->entrance_message = ST_YOU_RETURN_TO_THE_COUNTRYSIDE;
            flag_set(&result->flags, NO_FREE);
            exit_border = FALSE;
        break;
        case LEVEL_CRYPT:
            result->environment_name = ST_A_CRYPT;
            result->entrance_message = ST_YOU_DESCEND_INTO_A_CRYPT;
            result->exit_message = ST_YOU_ASCEND_FROM_A_CRYPT;
        break;
        case LEVEL_FINAL_QUEST:
            result->environment_name = ST_A_MYSTERIOUS_DUNGEON;
        break;
        case LEVEL_NORKRIN_ICE_PEAKS_HIDDEN_POOL:
            result->environment_name = ST_THE_ICE_PEAKS;
            result->race = NORKRIN;
            flag_set(&result->flags, OUTDOORS);
            exit_border = TRUE;
        break;
        case LEVEL_OCEAN_ISLE:
            result->environment_name = ST_A_SMALL_ISLAND_IN_THE_OCEAN;
            flag_set(&result->flags, OUTDOORS);
            exit_border = TRUE;
        break;
        case LEVEL_TAVERN:
            result->environment_name = ST_A_TAVERN;
            result->entrance_message = ST_YOU_ENTER_A_TAVERN;
            result->exit_message = ST_YOU_LEAVE_A_TAVERN;
            result->race = current_level->race;
        break;
        }
    /* All level files will be MAX_X by MAX_Y -- i.e. complete tables for */
    /* levels.  If for some reason we can't load all of a level, we consider */
    /* that a fatal error, and then exit.  */
    /* */
    /* I am being very picky about the files functioning properly.  I am */
    /* focusing on getting the files set up properly and then making sure */
    /* that they are not corrupt, not on robustness in having the game */
    /* function properly even if the files are corrupt.  */
    for(j=0;j<MAX_Y;++j)
        {
        if (!(fgets(this_row, MAX_X+2, infile)))
            {
#ifdef MAC_68k
            pmsg("Error: Premature end of file in file:");
            pmsg("");
            pmsg(filename);
            pmsg("");
            pmsg("Aborting...");
#else
            perror("Error: Premature end of file in file:");
            perror("");
            perror(filename);
            perror("");
            perror("Aborting...");
#endif
            exit(1);
            }
        else
            {
            for(i=0;i<MAX_X;++i)
                if (this_row[i] == 0) /* premature null termination */
                    {
#ifdef MAC_68k
                    pmsg("Error: Premature end of line in file:");
                    pmsg("");
                    pmsg(filename);
                    pmsg("");
                    pmsg("Line in file:");
                    pmsg("");
                    pmsg(this_row);
                    pmsg("");
                    pmsg("Aborting...");
#else
                    perror("Error: Premature end of line in file:");
                    perror("");
                    perror(filename);
                    perror("");
                    perror("Line in file:");
                    perror("");
                    perror(this_row);
                    perror("");
                    perror("Aborting...");
#endif
                    printf("\ni is %ld; j is %ld\n.", i, j);
                    exit(1);
                    }
                else
                    {
                    switch(level_type)
                        {
                        /* Different character interpretations for */
                        /* countryside and other levels */
                        case LEVEL_CATHEDRAL: case LEVEL_CRYPT: case
                          LEVEL_FINAL_QUEST: case
                          LEVEL_NORKRIN_ICE_PEAKS_HIDDEN_POOL: case
                          LEVEL_OCEAN_ISLE: case LEVEL_TAVERN:
                            switch(this_row[i])
                                {
                                case '.':
                                    change_space(&result->space[i][j],
                                      S_FLOOR, result->race);
                                break;
                                case '@':
                                    change_space(&result->space[i][j],
                                      S_FLOOR, result->race);
                                    if (level_type == LEVEL_FINAL_QUEST)
                                        change_space(&result->space[i][j],
                                          S_PEAK, result->race);
                                    game_variables->new_level_player_x = i;
                                    game_variables->new_level_player_y = j;
                                break;
                                case '#':
                                    change_space(&result->space[i][j],
                                      S_WALL, result->race);
                                break;
                                case '%':
                                    change_space(&result->space[i][j], S_FIRE,
                                      result->race);
                                    add_light_source_to_spot(result, i, j, 100);
                                break;
                                case '^':
                                    change_space(&result->space[i][j],
                                      S_PEAK, result->race);
                                break;
                                case '&':
                                    change_space(&result->space[i][j],
                                      S_FLOOR, result->race);
                                    if (trandom() % 2 == 0)
                                        add_item(&result->space[i][j].items,
                                          generate_item(result,
                                          DEVICE_NORMAL));
                                break;
                                case '(':
                                    change_space(&result->space[i][j],
                                      S_GREEN_TREE, result->race);
                                break;
                                case '-':
                                    change_space(&result->space[i][j],
                                      S_CLOSED_DOOR, result->race);
                                break;
                                case '|':
                                    change_space(&result->space[i][j],
                                      S_OPEN_DOOR, result->race);
                                break;
                                case '\'':
                                    change_space(&result->space[i][j],
                                      S_FLOOR, result->race);
                                    flag_set(&result->space[i][j].flags,
                                      S_CLOAKED_BY_MIST);
                                break;
                                case '"':
                                    if (level_type == LEVEL_OCEAN_ISLE)
                                        change_space(&result->space[i][j],
                                          S_SALT_WATER, result->race);
                                    else
                                        change_space(&result->space[i][j],
                                          S_WATER, result->race);
                                break;
                                case '<':
                                    change_space(&result->space[i][j],
                                      S_STAIRWELL_UP, result->race);
                                    if (level_type == LEVEL_CRYPT)
                                        {
                                        game_variables->new_level_player_x = i;
                                        game_variables->new_level_player_y = j;
                                        }
                                break;
                                case '?':
                                    if (level_type == LEVEL_FINAL_QUEST)
                                        change_space(&result->space[i][j],
                                          S_FINAL_TEST, result->race);
                                break;
                                case '1':
                                    if (level_type == LEVEL_CATHEDRAL)
                                        {
                                        change_space(&result->space[i][j],
                                          S_PEW, result->race);
                                        if (trandom() % 3 == 0)
                                            {
                                            this_creature = new_npc_generic(
                                              result->race);
                                            flag_reset(
                                              &this_creature->flags,
                                              C_WANDERING);
                                            add_creature(result, this_creature,
                                              i, j);
                                            }
                                        }
                                    else if (level_type == LEVEL_TAVERN)
                                        {
                                        change_space(&result->space[i][j],
                                          S_CHAIR, result->race);
                                        if (trandom() % 4 == 0)
                                            {
                                            this_creature = new_npc_generic(
                                              result->race);
                                            flag_reset(
                                              &this_creature->flags,
                                              C_WANDERING);
                                            add_creature(result, this_creature,
                                              i, j);
                                            }
                                        result->space[i][j].property =
                                          P_TAVERN_SEAT;
                                        flag_set(&result->space[i][j].flags,
                                          AUTOMATIC_EFFECT);
                                        }
                                    else
                                        change_space(&result->space[i][j],
                                          S_CHAIR, result->race);
                                break;
                                case '2':
                                    change_space(&result->space[i][j],
                                      S_BED, result->race);
                                break;
                                case '3':
                                    change_space(&result->space[i][j],
                                      S_WINDOW, result->race);
                                break;
                                case '4':
                                    change_space(&result->space[i][j],
                                      S_TABLE, result->race);
                                break;
                                case '5':
                                    change_space(&result->space[i][j], 
                                      S_STATUE, result->race);
                                    flag_reset(&result->space[i][j].flags,
                                      CAN_OCCUPY);
                                break;
                                case '6':
                                    change_space(&result->space[i][j],
                                      S_CABINET, result->race);
                                break;
                                case '7':
                                    change_space(&result->space[i][j],
                                      S_SHELVES, result->race);
                                break;
                                case '8':
                                    change_space(&result->space[i][j],
                                      S_FOUNTAIN, result->race);
                                break;
                                case '9':
                                    change_space(&result->space[i][j],
                                      S_ALTAR, result->race);
                                break;
                                case 'a': case 'b': case 'c': case 'd': case
                                  'e': case 'f': case 'g': case 'h': case 'i':
                                  case 'j': case 'k': case 'l': case 'm': case
                                  'n': case 'o': case 'p': case 'q': case 'r':
                                  case 's': case 't': case 'u': case 'v': case
                                  'w': case 'x': case 'y': case 'z': case 'A':
                                  case 'B': case 'C': case 'D': case 'E': case
                                  'F': case 'G': case 'H': case 'I': case 'J':
                                  case 'K': case 'L': case 'M': case 'N': case
                                  'O': case 'P': case 'Q': case 'R': case 'S':
                                  case 'T': case 'U': case 'V': case 'W': case
                                  'X': case 'Y': case 'Z':
                                    switch(level_type)
                                        {
                                        case LEVEL_CATHEDRAL:
                                            switch(this_row[i])
                                                {
                                                case 'b':
                                                    change_space(&result->space[i][j],
                                                      S_COLLECTION_BOX,
                                                      result->race);
                                                break;
                                                case 'c':
                                                    change_space(&result->space[i][j],
                                                      S_STAIRWELL_DOWN,
                                                      result->race);
                                                    result->space[i][j].
                                                      special->level_type
                                                      = LEVEL_CRYPT;
                                                    flag_set(&result->space[i]
                                                      [j].special->level_flags,
                                                      LOAD);
                                                break;
                                                case 'n':
                                                    change_space(&result->space
                                                      [i][j], S_STATUE,
                                                      result->race);
                                                    result->space[i][j].
                                                      space_description =
                                                      ST_A_BRONZE_STATUE_IN_A_NATIVITY_SCENE;
                                                break;
                                                case 'o':
                                                    change_space(&result->space[i][j],
                                                      S_FLOOR, result->race);
                                                    add_item(&(result->space[i][j].items),
                                                      generate_item(result, ORGAN));
                                                    flag_reset(&result->space[i][j].flags,
                                                      CAN_OCCUPY);
                                                break;
                                                case 'r':
                                                    change_space(&result->space[i][j],
                                                      S_FLOOR, result->race);
                                                    add_item(&result->space[i][j].items,
                                                      generate_item(result, RELIC));
                                                break;
                                                case 's':
                                                    change_space(&result->space[i][j],
                                                      S_SECRET_PASSAGE,
                                                      result->race);
                                                break;
                                                case 't':
                                                    change_space(&result->space[i][j],
                                                      S_PRAYER_TABLE,
                                                      result->race);
                                                    add_light_source_to_spot(
                                                      result, i, j, 10);
                                                break;
                                                case 'x':
                                                    change_space(&result->space[i][j],
                                                      S_AUTOMATIC_EXIT,
                                                      result->race);
                                                break;
                                                case 'w':
                                                    change_space(&result->space[i][j],
                                                      S_STAINED_GLASS_WINDOW,
                                                      result->race);
                                                break;
                                                case 'C':
                                                    change_space(&result->space[i][j],
                                                      S_FLOOR, result->race);
                                                    this_creature =
                                                      new_npc_generic(
                                                      result->race);
                                                    this_creature->extended_creature->
                                                      talk = TALK_CELEBRANT;
                                                    flag_reset(&this_creature->
                                                      flags, C_WANDERING);
                                                    add_creature(result,
                                                      this_creature, i, j);
                                                break;
                                                case 'I':
                                                    change_space(&result->space[i][j],
                                                      S_FLOOR, result->race);
                                                    this_creature =
                                                      new_npc_generic(
                                                      result->race);
                                                    this_creature->extended_creature->
                                                      talk = TALK_INTERCESSOR;
                                                    flag_reset(&this_creature->
                                                      flags, C_WANDERING);
                                                    add_creature(result,
                                                      this_creature, i, j);
                                                break;
                                                case 'P':
                                                    change_space(&result->space[i][j],
                                                      S_FLOOR, result->race);
                                                    if (trandom() % 6 == 0)
                                                        {
                                                        this_creature =
                                                          new_npc_generic(
                                                          result->race);
                                                        flag_reset(
                                                          &this_creature->
                                                          flags, C_WANDERING);
                                                        add_creature(result,
                                                          this_creature, i, j);
                                                        }
                                                break;
                                                }
                                        break;
                                        case LEVEL_CRYPT:
                                            switch(this_row[i])
                                                {
                                                case 't':
                                                    change_space(&result->space
                                                      [i][j], S_TOMB,
                                                      result->race);
                                                break;
                                                }
                                        break;
                                        case LEVEL_FINAL_QUEST:
                                            switch(this_row[i])
                                                {
                                                case 'i':
                                                    change_space(&result->space
                                                      [i][j], S_FLOOR,
                                                      result->race);
                                                    result->space[i][j].
                                                      property =
                                                      P_INSCRIPTION;
                                                break;
                                                case 'm':
                                                    change_space(&result->space
                                                      [i][j], S_PEAK,
                                                      result->race);
                                                    /*
                                                    change_space(&result->space
                                                      [i][j], S_FLOOR,
                                                      result->race);
                                                    */
                                                    flag_set(&result->space
                                                      [i][j].flags,
                                                      S_CLOAKED_BY_MIST);
                                                    if ((!(flag_check(
                                                      result->space[i][j].
                                                      flags, OCCUPIED))) &&
                                                      (trandom() % 150 == 0))
                                                        {
                                                        this_creature =
                                                          (LEVEL_1_ENCOUNTER +
                                                          (trandom() % 5));
                                                        make_hostile(
                                                          this_creature,
                                                          player);
                                                        add_creature(result,
                                                          this_creature, i, j);
                                                        }
                                                break;
                                                case 's':
                                                    change_space(&result->space
                                                      [i][j],
                                                      S_SECRET_PASSAGE,
                                                      result->race);
                                                    result->space[i][j].other =
                                                      S_FLOOR;
                                                break;
                                                }
                                        break;
                                        case
                                          LEVEL_NORKRIN_ICE_PEAKS_HIDDEN_POOL:
                                            switch(this_row[i])
                                                {
                                                case 'p':
                                                    change_space(&result->space
                                                      [i][j], S_SHALLOW_WATER,
                                                      result->race);
                                                    result->space[i][j].
                                                      property =
                                                      P_NORTHERN_POOL;
                                                break;
                                                }
                                        break;
                                        case LEVEL_OCEAN_ISLE:
                                            switch(this_row[i])
                                                {
                                                case 'b':
                                                    change_space(&result->space
                                                      [i][j], S_SHALLOW_WATER,
                                                      result->race);
                                                    flag_set(&result->space
                                                      [i][j].flags,
                                                      SALT_WATER);
                                                break;
                                                case 's':
                                                    change_space(&result->space
                                                      [i][j], S_FLOOR,
                                                      result->race);
                                                    flag_set(&result->space
                                                      [i][j].flags,
                                                      HAS_SECRET);
                                                    result->space[i][j].other =
                                                      S_PEARL_OF_GREAT_PRICE;
                                                break;
                                                }
                                        break;
                                        case LEVEL_TAVERN:
                                            switch(this_row[i])
                                                {
                                                case 'B':
                                                    this_creature =
                                                      new_npc_generic(
                                                      result->race);
                                                    flag_reset(&this_creature->
                                                      flags, C_WANDERING);
                                                    add_creature(result,
                                                      this_creature, i, j);
                                                break;
                                                case 'P':
                                                    if (trandom() % 4 == 0)
                                                        {
                                                        this_creature =
                                                          new_npc_generic(
                                                          result->race);
                                                        flag_reset(
                                                          &this_creature->
                                                          flags, C_WANDERING);
                                                        add_creature(result,
                                                          this_creature, i, j);
                                                        }
                                                break;
                                                case 'S':
                                                    this_creature =
                                                      new_npc_generic(
                                                      result->race);
                                                    this_creature->extended_creature->
                                                      talk = TALK_SERVER;
                                                    flag_reset(&this_creature->
                                                      flags, C_WANDERING);
                                                    add_creature(result,
                                                      this_creature, i, j);
                                                break;
                                                case 'b':
                                                    change_space(&result->
                                                      space[i][j],
                                                      S_INN_ENTRANCE,
                                                      result->race);
                                                break;
                                                case 't':
                                                    change_space(&result->
                                                      space[i][j], S_TORCH,
                                                      result->race);
                                                    add_light_source_to_spot(
                                                      result, i, j, 10);
                                                break;
                                                case 'x':
                                                    change_space(&result->space
                                                      [i][j],
                                                      S_AUTOMATIC_EXIT,
                                                      result->race);
                                                }
                                        break;
                                        }
                                break;
                                }
                        break;
                        case LEVEL_COUNTRYSIDE:
                            switch(this_row[i])
                                {
                                case '\"':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_OCEAN, result->race);
                                break;
                                case 'N':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_NORKRIN_PLAINS,
                                      result->race);
                                break;
                                case '2':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_NORKRIN_EAST_RIDGE_MOUNTAINS,
                                      result->race);
                                break;
                                case '1':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_NORKRIN_ICE_PEAKS,
                                      result->race);
                                break;
                                case '!':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_NORKRIN_ICE_PEAKS_HIDDEN_POOL,
                                      result->race);
                                break;
                                case 'C':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_MIDDLE_PLAINS,
                                      result->race);
                                break;
                                case '3':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_EAST_RIDGE_MOUNTAINS,
                                      result->race);
                                break;
                                case 'U':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_PLAINS,
                                      result->race);
                                break;
                                case 'Y':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_FOREST,
                                      result->race);
                                break;
                                case '4':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_FOG_VALLEY,
                                      result->race);
                                    flag_set(&result->space[i][j].special->level_flags, L_CLOAKED_BY_MIST);
                                break;
                                case '5':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_EAST_RIDGE_MOUNTAINS,
                                      result->race);
                                break;
                                case 'T':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_FOREST, result->race);
                                break;
                                case '&':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_EAST_RIDGE_MOUNTAINS,
                                      result->race);
                                    flag_set(&result->space[i][j].flags,
                                      HAS_SECRET);
                                    result->space[i][j].other =
                                      S_COUNTRYSIDE_CAVERN_ENTRANCE;
                                    /* The entrance to a set of caverns */
                                break;
                                case '$':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_OCEAN_HIDDEN_ISLE,
                                      result->race);
                                break;
                                case 'S':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_FOREST, result->race);
                                break;
                                case 'a':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_MILN;
                                break;
                                case 'b':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_KANS;
                                break;
                                case 'c':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_HOS;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'd':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_NUV;
                                break;
                                case 'e':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_ELHANNADAN;
                                break;
                                case 'f':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_TOL;
                                break;
                                case 'g':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_HIDDEN_MANSION,
                                      result->race);
                                break;
                                case 'h':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_VERN;
                                break;
                                case 'i':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_HISH;
                                break;
                                case 'j':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_KILNATRONA;
                                break;
                                case 'k':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_JEC_TOWN, result->race);
                                    result->space[i][j].special->other =
                                      JEC_TOWN_LUT;
                                break;
                                case 'l':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_TALMIRNATRII;
                                break;
                                case 'm':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_SANDRELLA;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'n':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_ELMIRTRA;
                                break;
                                case 'o':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_CANSOMILTRAMA;
                                break;
                                case 'p':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_CAPITELLO;
                                break;
                                case 'q':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_ELIRE;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'r':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_EHOLINA;
                                break;
                                case 's':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    flag_set(&result->space[i][j].special->
                                      level_flags, NO_FREE);
                                    flag_set(&result->space[i][j].special->
                                      level_flags, L_CLOAKED_BY_MIST);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_MISTRELLI;
                                break;
                                case 't':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_URVANOVESTILLI_CITY,
                                      result->race);
                                    result->space[i][j].special->other =
                                      URVANOVESTILLI_CITY_STANDRILNATRELLA;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, NO_FREE);
                                    game_variables->new_level_player_x = i;
                                    game_variables->new_level_player_y = j;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, HAS_MENTOR);
                                    /* Prepare starting city */
                                break;
                                case 'u':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_SALINA;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'v':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_FISRATIL;
                                break;
                                case 'w':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_RHALLTIR;
                                break;
                                case 'x':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_SANLI;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'y':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_DOSSE;
                                break;
                                case 'z':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_BORD;
                                break;
                                case 'A':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_KUNDER;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'B':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_LUZZ;
                                break;
                                case 'D':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_TIRNA;
                                break;
                                case 'E':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_RASTI;
                                break;
                                case 'F':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_ZUT;
                                break;
                                case 'G':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_VASLI;
                                break;
                                case 'H':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_YEDIDIA_VILLAGE,
                                      result->race);
                                    result->space[i][j].special->other =
                                      YEDIDIA_VILLAGE_SILNA;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'I':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_STORNGE;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'J':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_KORG;
                                break;
                                case 'K':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_RHAV;
                                break;
                                case 'L':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_TUZZNIR;
                                break;
                                case 'M':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_ZOLK;
                                break;
                                case 'O':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_WAZ;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'P':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_GORF;
                                break;
                                case 'Q':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_KANDRHI;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'R':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_GHIRTON;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'V':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      SHAL_VILLAGE_KASSILLA;
                                    flag_set(&result->space[i][j].special->
                                      level_flags, IS_SEASIDE);
                                break;
                                case 'W':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      SHAL_VILLAGE_SALN;
                                break;
                                case 'X':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      SHAL_VILLAGE_SIF;
                                break;
                                case 'Z':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      SHAL_VILLAGE_INGSA;
                                break;
                                case '7':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      SHAL_VILLAGE_NOLDI;
                                break;
                                case '8':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      SHAL_VILLAGE_MAC_68kEP;
                                break;
                                case '9':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_TUZ_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      TUZ_VILLAGE_DOZ;
                                break;
                                case '0':
                                    change_space(&result->space[i][j],
                                      S_COUNTRYSIDE_SHAL_VILLAGE, result->race);
                                    result->space[i][j].special->other =
                                      SHAL_VILLAGE_HESRI;
                                break;
                                }
                        break;
                        }
                    }
            }
        }
    switch(level_type)
        {
        case LEVEL_CATHEDRAL: case LEVEL_TAVERN:
            for(i = 0; i < MAX_X; ++i)
                for(j = 0; j < MAX_Y; ++j)
                    {
                    flag_set(&result->space[i][j].flags, DIMLY_LIT);
                    flag_set(&result->space[i][j].flags, PERM_DIMLY_LIT);
                    }
        break;
        }
    if (exit_border)
        {
        for(i=0;i<MAX_X;++i)
            {
            result->space[i][0].property = P_EXIT;
            flag_set(&result->space[i][0].flags, AUTOMATIC_EFFECT);
            flag_reset(&result->space[i][0].flags, EFFECT_BEFORE_MOVE);
            result->space[i][0].special = new_special();
            result->space[i][MAX_Y - 1].property = P_EXIT;
            flag_set(&result->space[i][MAX_Y - 1].flags, AUTOMATIC_EFFECT);
            flag_reset(&result->space[i][MAX_Y - 1].flags, EFFECT_BEFORE_MOVE);
            result->space[i][MAX_Y - 1].special = new_special();
            }
        for(j=0;j<MAX_Y;++j)
            {
            result->space[0][j].property = P_EXIT;
            flag_set(&result->space[0][j].flags, AUTOMATIC_EFFECT);
            flag_reset(&result->space[0][j].flags, EFFECT_BEFORE_MOVE);
            result->space[0][j].special = new_special();
            result->space[MAX_X - 1][j].property = P_EXIT;
            flag_set(&result->space[MAX_X - 1][j].flags, AUTOMATIC_EFFECT);
            flag_reset(&result->space[MAX_X - 1][j].flags, EFFECT_BEFORE_MOVE);
            result->space[MAX_X - 1][j].special = new_special();
            }
        }
    fclose(infile);
    return result;
    }

void change_space(struct space_struct *space, long identity, long race)
    {
    if (!(space->special == NULL))
        {
        free_special(space->special);
        space->special=NULL;
        }
    prepare_space(space, identity, race);
    }
