/***************************************************************************/
/* The Minstrel's Song is (c) 1998-2005 by Jonathan Hayward.  See the file */
/* lib/copyrigh.txt for a copyright and license statement.                 */
/***************************************************************************/

#include "tms.h"
#include <math.h>

#ifdef CODE_SEGMENT
#pragma segment UTIL4_C
#endif

long attribute_adjust(long amount, long attribute)
    {
    return max(1, (long) ((double) amount * pow(2.0, ((double) attribute) /
      10.0)));
    }

void find_and_write_on_item(struct small_creature *c)
    {
    struct small_item *itarget;
    if (c->extended_creature != NULL)
        {
        if (has_item_id(c->extended_creature->wear_wield, PEN, 0))
            {
            itarget = get_item(c, EXTRACT_RECURSIVE);
            if (itarget != NULL)
                {
                if (flag_check(itarget->flags, CAN_WRITE_ON))
                    {
                    take_time(c, 5L * MINUTE);
                    if (skill_check(c,
                      c->extended_creature->skills->av[READING_WRITING], DIFFICULT,
                      SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE) > 0)
                        {
                        if (flag_check(itarget->flags2, I_WRITE_ON_LOG))
                            write_on_log();
                        else
                            write_on_item(&itarget->writing);
                        }
                    else
                        pmsg("You can't concentrate to write anything.");
                    }
                else
                    pmsg("You can't write on that.");
                }
            }
        else
            pmsg("You don't have anything in your hand to write with.");
        }
    }

void append_clist(struct clist **c1, struct clist *c2)
    {
    struct clist *traverse;
    if (*c1 == NULL)
        *c1 = c2;
    else if ((*c1)->next == NULL)
        (*c1)->next = c2;
    else
        {
        for(traverse = *c1; (!(traverse->next == NULL)); traverse =
          traverse->next)
            my_noop();
        traverse->next = c2;
        }
    }

void append_ilist(struct ilist **i1, struct ilist *i2)
    {
    struct ilist *traverse;
    if (*i1 == NULL)
        *i1 = i2;
    else if ((*i1)->next == NULL)
        (*i1)->next = i2;
    else
        {
        for(traverse = *i1; (!(traverse->next == NULL)); traverse =
          traverse->next)
            my_noop();
        traverse->next = i2;
        }
    }

void antidote(struct small_creature *c1, struct small_creature *c2)
    {
    long medical_skill, success;
    if ((c1 != NULL) && (c2 != NULL))
        {
        if (c1->extended_creature != NULL)
            {
            medical_skill = c1->extended_creature->skills->av[MEDICINE];
            find_and_remove_id(&c1->extended_creature->pack, ANTIDOTE);
            }
        else
            medical_skill = 0;
        if (c2->extended_creature != NULL)
            success = min(c2->extended_creature->stati->time[STATUS_HEALABLE_ANTIDOTE],
             skill_check(c1, medical_skill, MODERATE,
              SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE));
        else
            success = min(c2->wounded,
             skill_check(c1, medical_skill, MODERATE,
              SKILL_CHECK_CRITICAL_SUCCESS_GAIN_EXPERIENCE));
        if (success > 0)
            healing(c2, success);
        }
    }

/* Decrease an item's number, and modify stats accordingly.  Return value
indicates that it was used up (TRUE).  */

static long decrease_number(struct small_item *i)
    {
    i->uses = (i->uses * (i->number - 1))/i->number;
    i->value = (i->value * (i->number - 1))/i->number;
    i->weight = (i->weight * (i->number - 1))/i->number;
    --i->number;
    if (i->number == 0)
        return TRUE;
    else
        return FALSE;
    }

/* Return status indicates whether the joining was successful.  */

static long join(struct small_item *i1, struct small_item **i2)
    {
    if ((i1 != NULL) && (*i2 != NULL))
        {
        if ((i1->identity == (*i2)->identity) && (i1->contents != NULL) &&
          ((*i2)->contents != NULL) && ((flag_check(i1->flags, FINE_QUALITY)) ==
          (flag_check((*i2)->flags, FINE_QUALITY))) && ((flag_check(i1->flags,
          POOR_QUALITY)) == (flag_check((*i2)->flags, POOR_QUALITY))))
            {
            i1->number += (*i2)->number;
            i1->uses += (*i2)->uses;
            i1->weight += (*i2)->weight;
            i1->value += (*i2)->value;
            free_item(*i2);
            *i2 = NULL;
            return TRUE;
            }
        else
            return FALSE;
        }
    else
        return FALSE;
    }

void add_item_join(struct ilist **il, struct small_item **i)
    {
    struct ilist *traverse;
    long my_continue;
    my_continue = TRUE;
    if (*il != NULL)
        for(traverse = *il; ((traverse != NULL) && (my_continue)); traverse =
          traverse->next)
        if (traverse->current != NULL)
            {
            my_continue = (!(join(traverse->current, i)));
            }
    if (*i != NULL) /* if the item was not joined  */
        add_item(il, *i);
    }

/* Returns whether it was successful.  */

long find_and_remove_id(struct ilist **il, long item_id)
    {
    struct ilist *traverse, *prev, *temp;
    long my_continue;
    if (*il != NULL)
        {
        if ((*il)->current->identity == item_id)
            {
            if (decrease_number((*il)->current))
                {
                temp = *il;
                *il = (*il)->next;
                free_item(temp->current);
                free(temp);
                }
            return TRUE;
            }
        else if ((*il)->next != NULL)
            {
            my_continue = TRUE;
            for(prev = traverse = *il; ((traverse != NULL) &&
              (my_continue)); traverse = traverse->next)
        if (traverse->current != NULL)
                {
                if (traverse->current->identity == item_id)
                    {
                    my_continue = FALSE;
                    if (decrease_number(traverse->current))
                        {
                        temp = traverse;
                        prev->next = traverse->next;
                        free_item(temp->current);
                        free(temp);
                        traverse = prev;
                        }
                    return TRUE;
                    }
                prev = traverse;
                }
            }
        }
    return FALSE;
    }

struct small_item *find_weapon(struct small_creature *c)
    {
    struct ilist *traverse;
    struct master_item real_weapon;
    if (c != NULL)
        if (c->extended_creature != NULL)
            if (c->extended_creature->wear_wield != NULL)
                for(traverse = c->extended_creature->wear_wield; traverse != NULL;
                  traverse = traverse->next)
        if (traverse->current != NULL)
                    {
                    real_weapon = *orig_items[traverse->current->identity];
                    if ((flag_check(real_weapon.flags2, IS_WEAPON)) &&
                      (flag_check(real_weapon.slots, PRIMARY_HAND)))
                        return traverse->current;
                    }
    return NULL;
    }

void encounter(long encounter_type, long success)
    {
    if (current_level == countryside)
        change_level(player->x, player->y);
    if (can_see(player, player->x, player->y))
        pmsg("You have an encounter!");
    encounter_internal(current_level, encounter_type, success);
    }

void human_encounter(long encounter_type, long success)
    {
    if (current_level == countryside)
        change_level(player->x, player->y);
    if (can_see(player, player->x, player->y))
        pmsg("You encounter some people!");
    human_encounter_internal(current_level, encounter_type, success);
    }

void go_to_sleep(struct small_creature *c)
    {
    flag_set(&c->flags, SLEEPING);
    flag_set(&c->flags, CANT_SEE);
    flag_set(&c->flags, INACTIVE);
    if (c->extended_creature != NULL)
        c->extended_creature->stati->time[STATUS_SLEEPING] =
          (adjustment_info[STATUS_TIRED].time_increment *
          (3 - adjustment_info[STATUS_TIRED].min) + c->extended_creature->stati->time
          [STATUS_TIRED])/3;
    if (c == player)
        pmsg("You fall asleep.");
    }

void wake_up(struct small_creature *c)
    {
    if (flag_check(c->flags, SLEEPING))
        {
        flag_reset(&c->flags, SLEEPING);
        flag_reset(&c->flags, CANT_SEE);
        flag_reset(&c->flags, INACTIVE);
        if (c == player)
            {
            pmsg("You wake up.");
            flag_reset(&game_variables->flags, PLAYER_SLEEPING_UNTIL_SUNRISE);
        update_light_sources();
            }
        }
    }

/* Call whatever routine is being debugged.  */

void debug(void)
    {
    char player_input[256];
    long quest_flag, i, j, k, item_id;
    if (flag_check(game_variables->flags, DEBUG_MODE))
        {
#ifdef GTK
        switch(new_tgetc("Debugging commands:\n\n\ta: gain massive experience\n\tb: gain 1000 gold\n\tc: set quest flag\n\td: clear negative stati\n\te: illuminate current level\n\tf: make current level known\n\tg: greatly increase faith\n\th: change current quest\n\ti: acquire an item\n\tj: complete current quest\n\tk: run mentor function\n\tl: run routine currently being tested"))
#else
        window_clear();
        fs_pmsg("Debugging commands:");
        fs_pmsg("");
        fs_pmsg("a: gain massive experience");
        fs_pmsg("b: gain 1000 gold");
        fs_pmsg("c: set quest flag");
        fs_pmsg("d: clear negative stati");
        fs_pmsg("e: illuminate current level");
        fs_pmsg("f: make current level known");
        fs_pmsg("g: greatly increase faith");
        fs_pmsg("h: change current quest");
        fs_pmsg("i: acquire an item");
        fs_pmsg("j: complete current quest");
        fs_pmsg("k: run mentor function");
        fs_pmsg("l: run routine currently being tested");
        switch(tgetc())
#endif
            {
            case 'a':
                gain_ep(player, 1000000L);
            break;
            case 'b':
                player->extended_creature->au += 1000;
                calc_ready(player);
            break;
            case 'c':
                window_clear();
                fs_pmsg("");
#ifdef GTK
                new_tgets("Please enter the value you would like to have bitwise or-ed with the quests record.", player_input, NULL);
#else
                fs_pmsg("Please enter the value you would like to have bitwise or-ed with the quests");
                fs_pmsg("record.");
                tgets(player_input);
#endif
#ifdef MAC_68k
                quest_flag = atoi(player_input);
#else
                sscanf(player_input, "%ld", &quest_flag);
#endif
                flag_set(&game_variables->quests_completed, quest_flag);
            break;
            case 'd':
                player->extended_creature->stati->time[STATUS_ALCOHOL] = 0;
                player->extended_creature->stati->time[STATUS_COLD] = 0;
                player->extended_creature->stati->time[STATUS_FATIGUED] = 0;
                player->extended_creature->stati->time[STATUS_HUNGRY] = 0;
                player->extended_creature->stati->time[STATUS_LOST] = 0;
                player->extended_creature->stati->time[STATUS_PASSED_OUT] = 0;
                player->extended_creature->stati->time[STATUS_SLEEPING] = 0;
                player->extended_creature->stati->time[STATUS_STINKY] = 0;
                player->extended_creature->stati->time[STATUS_THIRSTY] = 0;
                player->extended_creature->stati->time[STATUS_TIRED] = 0;
                player->extended_creature->stati->time[STATUS_WOUNDED] = 0;
                calc_ready(player);
            break;
            case 'e':
                for(i = 0; i < MAX_X; ++i)
                    for(j = 0; j < MAX_Y; ++j)
                        flag_set(&current_level->space[i][j].flags,
                          CURR_ILLUMINATED|PERM_ILLUMINATED);
            break;
            case 'f':
                for(i = 0; i < MAX_X; ++i)
                    for(j = 0; j < MAX_Y; ++j)
                        flag_set(&current_level->space[i][j].flags, KNOWN);
                        if (flag_check(current_level->space[i][j].flags,
                          HAS_SECRET))
                            {
                            discover_secret(i, j);
                            flag_set(&current_level->space[i][j].flags, KNOWN);
                            }
                        for(k = 0; k < NUM_LOCATION_TYPES; ++k)
                            if (flag_check(current_level->flags, IS_CITY))
                                if (flag_check(
                                  current_level->space[i][j].flags,
                                  OTHER_IS_LOCATION_TYPE) &&
                                  (current_level->space[i][j].other
                                  == current_level->location[k].identity))
                                    if (!(flag_check(current_level->
                                      location[k].flags,
                                      LOCATION_KNOWN)))
                                        {
                                        flag_set(
                                          &current_level->location[k].
                                          flags, LOCATION_KNOWN);
                                        current_level->location[k].x = i;
                                        current_level->location[k].y = j;
                                        }
                            update_window();
            break;
            case 'g':
                player->extended_creature->faith->faith += 1000L;
            break;
            case 'h':
                window_clear();
                fs_pmsg("");
#ifdef GTK
                new_tgets("Please enter the number of the quest.", player_input, NULL);
#else
                fs_pmsg("Please enter the number of the quest.");
                tgets(player_input);
#endif
#ifdef MAC_68k
                game_variables->current_quest = atoi(player_input);
#else
                sscanf(player_input, "%ld", &game_variables->current_quest);
#endif
            break;
            case 'i':
                item_id = item_parse(I_ALL_ITEMS);
                if (item_id != NONE)
                    acquire_item(player, item_id);
            break;
            case 'j':
                game_variables->quests_completed |=
                  game_variables->current_quest;
            break;
            case 'k':
                p_mentor();
            break;
            case 'l':
                write_on_roster();
            break;
            }
        fs_anykey();
        window_clear();
        }
    else
        {
#ifdef GTK
        new_tgets("Please enter the debugging password.", player_input, NULL);
#else
        pmsg("Please enter the debugging password.");
        tgets(player_input);
#endif
        if (!(strcmp(player_input, "50mwrx34")))
            {
            flag_set(&game_variables->flags, DEBUG_MODE);
            pmsg("Debug mode set.");
            debug();
            }
        else
            pmsg("Ernk.  Sorry.  Thank you for playing, but no.");
        }
    }

/* Fatigue a creature for a specified time of intense exercise.  */

void fatigue(struct small_creature *c, long units)
    {
    long result, old_fatigued, old_hungry, old_thirsty, old_tired;
    /* Disabled 20 Mar 03 by Jonathan Hayward because fatigue seemed not to
     * adequately wear off, and to be nicer to the player. */
    return;
    old_fatigued = status_check(c, STATUS_FATIGUED);
    old_hungry = status_check(c, STATUS_HUNGRY);
    old_thirsty = status_check(c, STATUS_THIRSTY);
    old_tired = status_check(c, STATUS_TIRED);
    if ((c->extended_creature != NULL) && (!(status_check(c,
      STATUS_MIRACLE_RESTORATION))))
        {
        result = skill_check(c, max(c->extended_creature->skills->av[ENDURANCE],
          c->extended_creature->skills->bv[ENDURANCE]), DIFFICULT,
          0);
        c->extended_creature->stati->time[STATUS_FATIGUED] += units * 10 * (1 - result);
        c->extended_creature->stati->time[STATUS_FATIGUED] =
          max(c->extended_creature->stati->time[STATUS_FATIGUED],
          adjustment_info[STATUS_FATIGUED].time_increment *
          adjustment_info[STATUS_FATIGUED].min);
        c->extended_creature->stati->time[STATUS_COLD] -= units * 5;
        c->extended_creature->stati->time[STATUS_COLD] =
          max(c->extended_creature->stati->time[STATUS_COLD],
          adjustment_info[STATUS_COLD].time_increment *
          adjustment_info[STATUS_COLD].min);
        c->extended_creature->stati->time[STATUS_HUNGRY] += units * 10;
        c->extended_creature->stati->time[STATUS_THIRSTY] += units * 10;
        c->extended_creature->stati->time[STATUS_TIRED] += units * 10 * (1 - result);
        c->extended_creature->stati->time[STATUS_TIRED] =
          max(c->extended_creature->stati->time[STATUS_TIRED],
          adjustment_info[STATUS_TIRED].time_increment *
          adjustment_info[STATUS_TIRED].min);
        }
    if (c == player)
        {
        if (old_fatigued != status_check(c, STATUS_FATIGUED))
            status_message(c, STATUS_FATIGUED);
        if (old_hungry != status_check(c, STATUS_HUNGRY))
            status_message(c, STATUS_HUNGRY);
        if (old_thirsty != status_check(c, STATUS_THIRSTY))
            status_message(c, STATUS_THIRSTY);
        if (old_tired != status_check(c, STATUS_TIRED))
            status_message(c, STATUS_TIRED);
        }
    }

void fill_with_water(struct small_item *i)
    {
    long num_drinks;
    struct small_item *water;
    if (i != NULL)
        {
        if ((flag_check(i->flags, CONTAINER)) && (flag_check(i->flags,
          HOLD_LIQUID)))
            {
            if (i->contents == NULL)
                {
                if (orig_items[i->identity]->space > orig_items[WATER]->size)
                    {
                    num_drinks = orig_items[i->identity]->space /
                      orig_items[WATER]->size;
                    water = generate_item(current_level, WATER);
                    multiply_item(water, num_drinks);
                    insert_internal(current_level, water, i);
                    pmsg("You fill it...");
                    }
                else
                    pmsg("It's too small to be worth filling.");
                }
            else
                pmsg("You have to empty it first.");
            }
        else
            pmsg("That is not an appropriate container for liquids.");
        }
    }

void multiply_item(struct small_item *i, long number)
    {
    if (i != NULL)
        {
        i->number *= number;
        i->weight *= number;
        i->value *= number;
        i->uses *= number;
        }
    }

void find_and_remove_item(struct small_item *i, struct ilist **il)
    {
    struct ilist *traverse, *prev, *temp;
    if (*il != NULL)
        {
        if ((*il)->current == i)
            {
            temp = *il;
            *il = (*il)->next;
            free(temp);
            }
        else
            {
            for(prev = traverse = *il; (traverse != NULL); traverse =
              traverse->next)
        if (traverse->current != NULL)
                {
                if (traverse->current == i)
                    {
                    temp = traverse;
                    prev->next = traverse->next;
                    free(temp);
                    traverse = prev;
                    }
                if (traverse->current->contents != NULL)
                    find_and_remove_item(i, &traverse->current->contents);
                prev = traverse;
                }
            }
        }
    }

long found_in_distribution(long unit_id, long list_id, long type)
    {
    long success, i, distribution[NUM_DIST_SLOTS][3];
    switch(type)
        {
        case DISTRIBUTION_CREATURE:
            switch(list_id)
                {
                case JEC_ENCOUNTER:
                    distribution_copy(jec_encounter, distribution);
                break;
                case NORKRIN_ENCOUNTER:
                    distribution_copy(norkrin_encounter, distribution);
                break;
                case SHAL_ENCOUNTER:
                    distribution_copy(shal_encounter, distribution);
                break;
                case TUZ_ENCOUNTER:
                    distribution_copy(tuz_encounter, distribution);
                break;
                case URVANOVESTILLI_ENCOUNTER:
                    distribution_copy(urvanovestilli_encounter, distribution);
                break;
                case YEDIDIA_ENCOUNTER:
                    distribution_copy(yedidia_encounter, distribution);
                break;
                case LEVEL_1_ENCOUNTER:
                    distribution_copy(level_1_encounter, distribution);
                break;
                case LEVEL_2_ENCOUNTER:
                    distribution_copy(level_2_encounter, distribution);
                break;
                case LEVEL_3_ENCOUNTER:
                    distribution_copy(level_3_encounter, distribution);
                break;
                case LEVEL_4_ENCOUNTER:
                    distribution_copy(level_4_encounter, distribution);
                break;
                case LEVEL_5_ENCOUNTER:
                    distribution_copy(level_5_encounter, distribution);
                break;
                case LAND_SPECIAL_ENCOUNTER:
                    distribution_copy(land_special_encounter, distribution);
                break;
                case SEA_ENCOUNTER:
                    distribution_copy(sea_encounter, distribution);
                break;
                case SEA_SPECIAL_ENCOUNTER:
                    distribution_copy(sea_special_encounter, distribution);
                break;
                default:
                    /*
                    pmsg("found_in_distribution called with inappropriate list id");
                    */
                    return 0;
                break;
                }
        break;
        case DISTRIBUTION_ITEM:
            switch(list_id)
                {
                case ARMOR_FINE: case ARMOR_NORMAL: case ARMOR_NORMAL_FINE:
                    distribution_copy(armor, distribution);
                break;
                case DEVICE_FINE: case DEVICE_NORMAL: case DEVICE_NORMAL_FINE:
                    distribution_copy(device, distribution);
                break;
                case FURNITURE_FINE: case FURNITURE_NORMAL: case
                  FURNITURE_NORMAL_FINE:
                    distribution_copy(furniture, distribution);
                break;
                case LIQUID_FINE: case LIQUID_NORMAL: case LIQUID_NORMAL_FINE:
                    distribution_copy(liquid, distribution);
                break;
                case NATURAL_FINE: case NATURAL_NORMAL: case
                  NATURAL_NORMAL_FINE:
                    distribution_copy(natural, distribution);
                break;
                case NON_DEVICE_EQUIPMENT_FINE: case
                  NON_DEVICE_EQUIPMENT_NORMAL: case
                  NON_DEVICE_EQUIPMENT_NORMAL_FINE:
                    distribution_copy(non_device_equipment, distribution);
                break;
                case NON_EQUIPMENT_TREASURE_FINE: case
                  NON_EQUIPMENT_TREASURE_NORMAL: case
                  NON_EQUIPMENT_TREASURE_NORMAL_FINE:
                    distribution_copy(non_equipment_treasure, distribution);
                break;
                case SUPPLIES_FINE: case SUPPLIES_NORMAL: case
                  SUPPLIES_NORMAL_FINE:
                    distribution_copy(supplies, distribution);
                break;
                case WEAPON_FINE: case WEAPON_NORMAL: case WEAPON_NORMAL_FINE:
                    distribution_copy(weapon, distribution);
                break;
                case BUILDING_ITEM_FINE: case BUILDING_ITEM_NORMAL: case
                  BUILDING_ITEM_NORMAL_FINE:
                    distribution_copy(building_item, distribution);
                break;
                case DUNGEON_ITEM_FINE: case DUNGEON_ITEM_NORMAL: case
                  DUNGEON_ITEM_NORMAL_FINE:
                    distribution_copy(dungeon_item, distribution);
                break;
                case OUTDOORS_ITEM_FINE: case OUTDOORS_ITEM_NORMAL: case
                  OUTDOORS_ITEM_NORMAL_FINE:
                    distribution_copy(outdoors_item, distribution);
                break;
                case TINY_ITEM_FINE: case TINY_ITEM_NORMAL: case
                  TINY_ITEM_NORMAL_FINE:
                    distribution_copy(tiny_item, distribution);
                break;
                case BOOKSTORE_ITEM_FINE: case BOOKSTORE_ITEM_NORMAL:
                  case BOOKSTORE_ITEM_NORMAL_FINE:
                    distribution_copy(bookstore_item, distribution);
                break;
                case HERBALIST_ITEM_FINE: case HERBALIST_ITEM_NORMAL:
                  case HERBALIST_ITEM_NORMAL_FINE:
                    distribution_copy(herbalist_item, distribution);
                break;
                case JEWELLER_ITEM_FINE: case JEWELLER_ITEM_NORMAL:
                  case JEWELLER_ITEM_NORMAL_FINE:
                    distribution_copy(jeweller_item, distribution);
                break;
                default:
                    pmsg("found_in_distribution called with inappropriate list id");
                    return 0;
                break;
                }
        break;
        default:
            pmsg("found_in_distribution called with inappropriate distribution id");
            return 0;
        break;
        }
    success = 0;
    for(i = 0; i < NUM_DIST_SLOTS; ++i)
        success |= ((distribution[i][0] == unit_id) || (distribution[i][1] ==
          unit_id));
    return success;
    }

void distribution_copy(long source[NUM_DIST_SLOTS][3], long
  destination[NUM_DIST_SLOTS][3])
    {
    long i, j;
    for(i = 0; i < NUM_DIST_SLOTS; ++i)
        for(j = 0; j < 3; ++j)
            destination[i][j] = source[i][j];
    }

void hostility_check(struct level *l, struct small_creature *c)
    {
    struct clist *traverse;
    if (!(flag_check(c->flags, HIDDEN)))
        if (l->creatures != NULL)
            for(traverse=l->creatures;(!(traverse==NULL));
              traverse=traverse->next)
        if (traverse->current != NULL)
                if (!(is_following(traverse->current, c)))
                    if ((trandom()%100 < master_creatures[traverse->current->
                      identity]->chance_hostility) && (traverse->current != c))
                        make_hostile(traverse->current, c);
    }

#ifdef GTK
void set_options(void)
    {
    /*
    new_set_options_internal("You are now setting options for the game.  For each option you have two\nchoices: keep the option as it is, or change it.\nAsk before pickup: If this variable is set, the game will ask you if you want\n to pick something up, when you run across it.  If it is not set, you will\nautomatically pick up whatever you run across.  (You can still always drop what\n you don't want.", &game_variables->flags, ASK_PICKUP);
    new_set_options_internal("Combat happy: This option controls whether your character will attack a creature\nwhich is not hostile, or merely swap places with it, when you move over a\ncreature.  It is probably good to set this flag when hunting and in some other\ncases, but generally to have it be reset.", &player->flags, BELLICOSE);
    fs_anykey();
    set_options_internal(&player->flags, BELLICOSE);
    */
    }

#else

void set_options(void)
    {
    window_clear();
    fs_pmsg("");
    fs_pmsg("You are now setting options for the game.  For each option you have three");
    fs_pmsg("choices: (s)et the option, (r)eset the option, or (p)ass over it.  Set");
    fs_pmsg("chooses the option and makes it true, reset rejects the option and makes it");
    fs_pmsg("false, and passing over an option leaves it as it is.");
    fs_anykey();
    fs_pmsg("");
    fs_pmsg("Prompt for messages: This option controls whether, when there are enough");
    fs_pmsg("messages printed since you last hit a key, the game prompts for you to hit");
    fs_pmsg("'space' before printing any more.  If this option is reset, some messages may");
    fs_pmsg("scroll by before you can read them.");
    fs_anykey();
    set_options_internal(&game_variables->flags, MESSAGES_PROMPT);
    fs_pmsg("");
    fs_pmsg("Ask before pickup: If this variable is set, the game will ask you if you want");
    fs_pmsg("to pick something up, when you run across it.  If it is not set, you will");
    fs_pmsg("automatically pick up whatever you run across.  (You can still always drop what");
    fs_pmsg("you don't want.");
    fs_anykey();
    set_options_internal(&game_variables->flags, ASK_PICKUP);
    fs_pmsg("");
    fs_pmsg("Maximum refresh: This controls whether the game updates the screen more");
    fs_pmsg("frequently, or less frequently.");
    fs_anykey();
    set_options_internal(&game_variables->flags, MAX_REFRESH);
    fs_pmsg("");
    fs_pmsg("Player at center: This option controls whether the game always keeps the player");
    fs_pmsg("at the center of the screen, or whether it lets him move about and only shifts");
    fs_pmsg("the perspective when the player is near the edge of the screen.  (Setting the");
    fs_pmsg("option keeps the player at the center.");
    fs_anykey();
    fs_pmsg("");
    set_options_internal(&game_variables->flags, PLAYER_AT_CENTER);
    fs_pmsg("Negative confirm: When you say that you want to do something which the game");
    fs_pmsg("regards as dangerous, the game will be more discouraging in asking you if you");
    fs_pmsg("really want to do that.  This is recommended for novice players.");
    fs_anykey();
    set_options_internal(&game_variables->flags, NEGATIVE_CONFIRM);
    fs_pmsg("");
    fs_pmsg("Confirm when in doubt: If you do something which the game is not sure of, this");
    fs_pmsg("option will have the game prompt for confirmation.  (It is related to the");
    fs_pmsg("previous option, but not the same.");
    fs_anykey();
    set_options_internal(&game_variables->flags, CONFIRM_WHEN_IN_DOUBT);
    fs_pmsg("");
    fs_pmsg("Free unneeded levels: The Minstrel's Song can forget all but a few levels when");
    fs_pmsg("the player leaves them, in order to save memory.  If this option is set, The");
    fs_pmsg("Minstrel's Song will take large amounts of memory and destroy and re-generate");
    fs_pmsg("levels frequently.");
    fs_anykey();
    set_options_internal(&game_variables->flags, FREE_UNNEEDED_LEVELS);
    fs_pmsg("");
    fs_pmsg("Always display staircase: This will enable you to always see a staircase, even");
    fs_pmsg("if there is an item, creature, etc. on top of it.  This can be very useful in");
    fs_pmsg("dungeons.");
    fs_anykey();
    set_options_internal(&game_variables->flags, ALWAYS_DISPLAY_STAIRCASES);
    fs_pmsg("");
    fs_pmsg("Bellicose: This option controls whether your character will attack a creature");
    fs_pmsg("which is not hostile, or merely swap places with it, when you move over a");
    fs_pmsg("creature.  It is probably good to set this flag when hunting and in some other");
    fs_pmsg("cases, but generally to have it be reset.");
    fs_anykey();
    set_options_internal(&player->flags, BELLICOSE);
        /*
    fs_pmsg("Save file compression level: this is an integer from 1 to 9, which controls how");
    fs_pmsg("much time the game spends compressing save files.  Lower numbers correspond");
    fs_pmsg("to faster compression and larger files; higher numbers to slower compression");
    fs_pmsg("and smaller files.  The default is 6.  Change the save file compression");
    fs_pmsg("level?");
    if (confirm_silent_internal())
        {
        window_clear();
        pmsg("New save file compression level:");
        if (tgets(player_input))
            {
            sscanf(player_input, "%ld",
              &game_variables->save_file_compression_level);
            game_variables->save_file_compression_level =
              min(max(game_variables->save_file_compression_level, 1), 9);
            }
        else
            pmsg("Aborted.");
        }
        */
    }

#endif

void set_options_internal(long *variable, long flags)
    {
    char response;
    fs_pmsg("");
    if (flag_check(*variable, flags))
        fs_pmsg("This option is currently set.");
    else
        fs_pmsg("This option is not currently set.");
    /* ## */
    fs_pmsg("(s)et the option, (r)eset, or (p)ass over it, leaving it unchanged?");
    response = tgetc();
    while ((response != 's') && (response != 'r') && (response != 'p'))
        {
        fs_pmsg("Please hit 's', 'r', or 'p'.");
        response = tgetc();
        }
    switch(response)
        {
        case 's':
            flag_set(variable, flags);
            pmsg("Set.");
        break;
        case 'r':
            flag_reset(variable, flags);
            pmsg("Reset.");
        break;
        case 'p':
            pmsg("Passed over.");
        break;
        default:
            pmsg("Error: exceptional response permitted in set_options_internal");
        break;
        }
    fs_anykey();
    window_clear();
    }

long has_item_id(struct ilist *il, long item_id, long flags)
    {
    long success;
    struct ilist *traverse;
    success = FALSE;
    if (il != NULL)
        for(traverse = il; traverse != NULL; traverse = traverse->next)
        if (traverse->current != NULL)
            {
            if (traverse->current->identity == item_id)
                success = TRUE;
            if (flag_check(flags, EXTRACT_RECURSIVE))
                if (traverse->current->contents != NULL)
                    success |= has_item_id(traverse->current->contents,
                      item_id, flags);
            }
    return success;
    }

void add_amount(struct small_creature *c, long amount)
    {
    long curr_price;
    if (c->extended_creature != NULL)
        {
/*
 *
 * curr_price is the remaining amount of money to be paid.  Shops always take
 * the lowest coin which the player has.
 *
 */
        curr_price = amount;
        c->extended_creature->au += curr_price/AU;
        curr_price = curr_price % AU;
        c->extended_creature->el += curr_price/EL;
        curr_price = curr_price % EL;
        c->extended_creature->si += curr_price/SI;
        curr_price = curr_price % SI;
        c->extended_creature->cu += curr_price/CU;
        curr_price = curr_price % CU;
        }
    }

long deduct_amount(struct small_creature *c, long amount)
    {
    long curr_price;
    if (c->extended_creature != NULL)
        {
        if (amount > (c->extended_creature->au * AU +
          c->extended_creature->el * EL + c->extended_creature->si * SI +
          c->extended_creature->cu * CU))
            return 0;
        else
            {
/*
 *
 * curr_price is the remaining amount of money to be paid.  Shops always take
 * the lowest coin which the player has.
 *
 */
            curr_price = amount;
            if (c->extended_creature->cu * CU > curr_price)
                {
                c->extended_creature->cu -= curr_price/CU;
                curr_price = curr_price % CU;
                }
            if (c->extended_creature->si * SI > curr_price)
                {
                c->extended_creature->si -= curr_price/SI;
                curr_price = curr_price % SI;
                }
            if (c->extended_creature->el * EL > curr_price)
                {
                c->extended_creature->el -= curr_price/EL;
                curr_price = curr_price % EL;
                }
            if (c->extended_creature->au * AU > curr_price)
                {
                c->extended_creature->au -= curr_price/AU;
                curr_price = curr_price % AU;
                }
            c->extended_creature->el -= curr_price/EL;
            curr_price = curr_price % EL;
            c->extended_creature->si -= curr_price/SI;
            curr_price = curr_price % SI;
            c->extended_creature->cu -= curr_price/CU;
            curr_price = curr_price % CU;
            return TRUE;
            }
        }
    return TRUE;
    }

/* Remove one instance of a creature from a clist.  */

void remove_creature_from_clist(struct small_creature *c, struct clist **cl)
    {
    struct clist *traverse, *prev, *temp;
    if (*cl != NULL)
        {
        if ((*cl)->current == c)
            {
            temp = *cl;
            *cl = (*cl)->next;
            free(temp);
            }
        else
            {
            for(prev = traverse = *cl; traverse != NULL;)
                {
                if (traverse->current == c)
                    {
                    prev->next = traverse->next;
                    free(traverse);
                    traverse = NULL;
                    }
                prev = traverse;
                if (traverse != NULL)
                    traverse = traverse->next;
                }
            }
        }
    }

void append_creature_to_clist(struct small_creature *c, struct clist **cl)
    {
    struct clist *traverse, *new_clist;
    new_clist = (struct clist *) calloc(1, sizeof(struct clist));
    new_clist->me = new_clist;
    new_clist->current = c;
    new_clist->next = NULL;
    if (*cl == NULL)
        *cl = new_clist;
    else if ((*cl)->next == NULL)
        (*cl)->next = new_clist;
    else
        {
        for(traverse = *cl; traverse->next != NULL; traverse = traverse->next)
            my_noop();
        traverse->next = new_clist;
        }
    }

void identify(struct small_item *i)
    {
    if (i != NULL)
        {
        flag_set(&i->flags, IDENTIFIED);
        if ((flag_check(i->flags, MAY_HAVE_COMPARTMENT)) && (flag_check(i->flags,
          CONTAINER)))
        pmsg("It has a hidden compartment; it may be opened etc. as a regular container.");
        }
    }
